"""Classes for standardizing molecules."""

from __future__ import annotations

from typing import Any, Optional

try:
    from typing import Self  # pylint: disable=no-name-in-module
except ImportError:
    from typing_extensions import Self

from rdkit import Chem
from rdkit.Chem.Scaffolds import MurckoScaffold as RDKIT_MurckoScaffold

from molpipeline.abstract_pipeline_elements.core import (
    MolToMolPipelineElement as _MolToMolPipelineElement,
)
from molpipeline.utils.molpipeline_types import OptionalMol, RDKitMol


class MurckoScaffold(_MolToMolPipelineElement):
    """MolToMol-PipelineElement which yields the Murcko-scaffold of a Molecule.

    The Murcko-scaffold is composed of all rings and the linker atoms between them.
    """

    def pretransform_single(self, value: RDKitMol) -> OptionalMol:
        """Extract Murco-scaffold of molecule.

        Parameters
        ----------
        value: RDKitMol
            RDKit molecule object which is transformed.

        Returns
        -------
        OptionalMol
            Murco-scaffold of molecule if possible, else InvalidInstance.
        """
        return RDKIT_MurckoScaffold.GetScaffoldForMol(value)


class MakeScaffoldGeneric(_MolToMolPipelineElement):
    """MolToMol-PipelineElement which sets all atoms to carbon and all bonds to single bond.

    Done to make scaffolds less speciffic.
    """

    def __init__(
        self,
        generic_atoms: bool = False,
        generic_bonds: bool = False,
        name: str = "MakeScaffoldGeneric",
        n_jobs: int = 1,
        uuid: Optional[str] = None,
    ) -> None:
        """Initialize MakeScaffoldGeneric.

        Note
        ----
        Making atoms or bonds generic will generate SMARTS strings instead of SMILES strings.
        This can be useful to search for scaffolds and substructures in data sets.
        Per default, the scaffold is returned as SMILES string with all atoms set to carbon and all bonds are single bonds.

        Parameters
        ----------
        generic_atoms: bool
            If True, all atoms in the molecule are set to generic atoms (*).
        generic_bonds: bool
            If True, all bonds in the molecule are set to any bonds.
        name: str
            Name of pipeline element.
        n_jobs: int
            Number of jobs to use for parallelization.
        uuid: Optional[str]
            UUID of pipeline element.

        Returns
        -------
        None
        """
        self.generic_atoms = generic_atoms
        self.generic_bonds = generic_bonds
        super().__init__(name=name, n_jobs=n_jobs, uuid=uuid)

    def pretransform_single(self, value: RDKitMol) -> OptionalMol:
        """Set all atoms to carbon and all bonds to single bond and return mol object.

        Parameters
        ----------
        value: RDKitMol
            RDKit molecule object which is transformed.

        Returns
        -------
        OptionalMol
            Molecule where all atoms are carbon and all bonds are single bonds.
            If transformation failed, it returns InvalidInstance.
        """
        scaffold = RDKIT_MurckoScaffold.MakeScaffoldGeneric(value)
        if self.generic_atoms:
            for atom in scaffold.GetAtoms():
                atom.SetAtomicNum(0)
        if self.generic_bonds:
            for bond in scaffold.GetBonds():
                bond.SetBondType(Chem.rdchem.BondType.UNSPECIFIED)
        return scaffold

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Get parameters of pipeline element.

        Parameters
        ----------
        deep: bool
            If True, return the parameters of the pipeline element.

        Returns
        -------
        dict[str, Any]
            Parameters of the pipeline element.
        """
        parent_params = super().get_params()
        if deep:
            parent_params.update(
                {
                    "generic_atoms": bool(self.generic_atoms),
                    "generic_bonds": bool(self.generic_bonds),
                }
            )
        else:
            parent_params.update(
                {
                    "generic_atoms": self.generic_atoms,
                    "generic_bonds": self.generic_bonds,
                }
            )
        return parent_params

    def set_params(self, **parameters: dict[str, Any]) -> Self:
        """Set parameters of pipeline element.

        Parameters
        ----------
        parameters: dict[str, Any]
            Parameters to set.

        Returns
        -------
        Self
            Pipeline element with set parameters.
        """
        param_copy = parameters.copy()
        generic_atoms = param_copy.pop("generic_atoms", None)
        generic_bonds = param_copy.pop("generic_bonds", None)
        if generic_atoms is not None:
            if not isinstance(generic_atoms, bool):
                raise ValueError("generic_atoms must be a boolean.")
            self.generic_atoms = generic_atoms
        if generic_bonds is not None:
            if not isinstance(generic_bonds, bool):
                raise ValueError("generic_bonds must be a boolean.")
            self.generic_bonds = generic_bonds
        return self



"""Init the module for mol2any pipeline elements."""

from molpipeline.mol2any.mol2bin import MolToBinary
from molpipeline.mol2any.mol2bool import MolToBool
from molpipeline.mol2any.mol2concatinated_vector import MolToConcatenatedVector
from molpipeline.mol2any.mol2inchi import MolToInchi, MolToInchiKey
from molpipeline.mol2any.mol2maccs_key_fingerprint import MolToMACCSFP
from molpipeline.mol2any.mol2morgan_fingerprint import MolToMorganFP
from molpipeline.mol2any.mol2net_charge import MolToNetCharge
from molpipeline.mol2any.mol2path_fingerprint import Mol2PathFP
from molpipeline.mol2any.mol2rdkit_phys_chem import MolToRDKitPhysChem
from molpipeline.mol2any.mol2smiles import MolToSmiles

__all__ = [
    "MolToBinary",
    "MolToConcatenatedVector",
    "MolToSmiles",
    "MolToMACCSFP",
    "MolToMorganFP",
    "MolToNetCharge",
    "Mol2PathFP",
    "MolToInchi",
    "MolToInchiKey",
    "MolToRDKitPhysChem",
    "MolToBool",
]

try:
    from molpipeline.mol2any.mol2chemprop import MolToChemprop  # noqa

    __all__.append("MolToChemprop")
except ImportError:
    pass



"""Test Murcko scaffold clustering estimator."""

from __future__ import annotations

import unittest

import numpy as np

from molpipeline.estimators import MurckoScaffoldClustering

SCAFFOLD_SMILES: list[str] = [
    "Cc1ccccc1",
    "Cc1cc(Oc2nccc(CCC)c2)ccc1",
    "c1ccccc1",
]

SCAFFOLD_SMILES_TEST_GENERIC: list[str] = SCAFFOLD_SMILES + ["c1ncccc1"]

LINEAR_SMILES: list[str] = ["CC", "CCC", "CCCN"]


class TestMurckoScaffoldClusteringEstimator(unittest.TestCase):
    """Test Murcko scaffold clustering estimator."""

    def test_murcko_scaffold_clustering_ignore(self) -> None:
        """Test Murcko scaffold clustering estimator."""

        for make_generic in [False, True]:
            estimator_ignore_linear: MurckoScaffoldClustering = (
                MurckoScaffoldClustering(
                    make_generic=make_generic,
                    n_jobs=1,
                    linear_molecules_strategy="ignore",
                )
            )

            # test basic scaffold-based clustering works as intended
            scaffold_cluster_labels = estimator_ignore_linear.fit_predict(
                SCAFFOLD_SMILES
            )
            expected_scaffold_labels = [1.0, 0.0, 1.0]

            self.assertEqual(estimator_ignore_linear.n_clusters_, 2)
            self.assertListEqual(
                list(scaffold_cluster_labels), expected_scaffold_labels
            )

            # test linear molecule handling. We expect the linear molecules to be ignored.
            input_smiles = SCAFFOLD_SMILES + LINEAR_SMILES
            cluster_labels = estimator_ignore_linear.fit_predict(input_smiles)
            nan_mask = np.isnan(cluster_labels)
            expected_nan_mask = [False, False, False, True, True, True]

            self.assertEqual(estimator_ignore_linear.n_clusters_, 2)
            self.assertListEqual(list(nan_mask), expected_nan_mask)
            self.assertListEqual(cluster_labels[~nan_mask].tolist(), [1.0, 0.0, 1.0])

    def test_murcko_scaffold_clustering_own_cluster(self) -> None:
        """Test Murcko scaffold clustering estimator."""

        for make_generic in [False, True]:
            # create new estimator with "own_cluster" strategy
            estimator_cluster_linear: MurckoScaffoldClustering = (
                MurckoScaffoldClustering(
                    make_generic=make_generic,
                    n_jobs=1,
                    linear_molecules_strategy="own_cluster",
                )
            )

            # test linear molecule handling. We expect the linear molecules to be clustered in the same cluster
            input_smiles = SCAFFOLD_SMILES + LINEAR_SMILES
            cluster_labels = estimator_cluster_linear.fit_predict(input_smiles)
            expected_cluster_labels = [1.0, 0.0, 1.0, 2.0, 2.0, 2.0]
            self.assertEqual(estimator_cluster_linear.n_clusters_, 3)
            self.assertListEqual(list(cluster_labels), expected_cluster_labels)

    def test_murcko_scaffold_clustering_generic(self) -> None:
        """Test Murcko scaffold clustering estimator with generic scaffold."""

        # test generic clustering makes a difference
        estimator: MurckoScaffoldClustering = MurckoScaffoldClustering(
            make_generic=True,
            n_jobs=1,
            linear_molecules_strategy="ignore",
        )

        scaffold_cluster_labels = estimator.fit_predict(SCAFFOLD_SMILES_TEST_GENERIC)
        expected_scaffold_labels = [1.0, 0.0, 1.0, 1.0]

        self.assertEqual(estimator.n_clusters_, 2)
        self.assertListEqual(list(scaffold_cluster_labels), expected_scaffold_labels)

        # test that without make_generic we get a different result
        estimator2: MurckoScaffoldClustering = MurckoScaffoldClustering(
            make_generic=False,
            n_jobs=1,
            linear_molecules_strategy="ignore",
        )

        scaffold_cluster_labels2 = estimator2.fit_predict(SCAFFOLD_SMILES_TEST_GENERIC)
        expected_scaffold_labels2 = [1.0, 0.0, 1.0, 2.0]

        self.assertEqual(estimator2.n_clusters_, 3)
        self.assertListEqual(list(scaffold_cluster_labels2), expected_scaffold_labels2)

    def test_murcko_scaffold_clustering(self) -> None:
        """Test Murcko scaffold clustering estimator for purely linear molecules."""
        test_smiles_failing = [
            "CCCCCCCCC(=CCCCCCCCC(=O)O)[N+](=O)[O-]",
            "CN(C)C(=O)C(C)(C)NC(=O)OC(C)(C)C",
        ]

        pipe = MurckoScaffoldClustering(
            make_generic=False,
            n_jobs=1,
            linear_molecules_strategy="ignore",
        )
        result = pipe.fit_predict(test_smiles_failing, None)
        self.assertTrue(np.isnan(result).all())



"""Test generation of RDKitPhysChem Descriptors."""

import unittest
from pathlib import Path

import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler

from molpipeline import ErrorFilter, FilterReinserter, Pipeline
from molpipeline.any2mol import SmilesToMol
from molpipeline.mol2any import MolToRDKitPhysChem
from molpipeline.mol2any.mol2rdkit_phys_chem import DEFAULT_DESCRIPTORS

data_path = Path(__file__).parents[2] / "test_data" / "mol_descriptors.tsv"


class TestMol2RDKitPhyschem(unittest.TestCase):
    """Unittest for MolToRDKitPhysChem, which calculates RDKitPhysChem Descriptors."""

    def test_descriptor_list(self) -> None:
        """Test if the descriptor list is as expected.

        Returns
        -------
        None
        """
        expected_descriptors = {
            "MaxAbsEStateIndex",
            "MaxEStateIndex",
            "MinAbsEStateIndex",
            "MinEStateIndex",
            "qed",
            "HeavyAtomMolWt",
            "ExactMolWt",
            "NumValenceElectrons",
            "NumRadicalElectrons",
            "MaxPartialCharge",
            "MinPartialCharge",
            "MaxAbsPartialCharge",
            "MinAbsPartialCharge",
            "FpDensityMorgan1",
            "FpDensityMorgan2",
            "FpDensityMorgan3",
            "BCUT2D_MWHI",
            "BCUT2D_MWLOW",
            "BCUT2D_CHGHI",
            "BCUT2D_CHGLO",
            "BCUT2D_LOGPHI",
            "BCUT2D_LOGPLOW",
            "BCUT2D_MRHI",
            "BCUT2D_MRLOW",
            "AvgIpc",
            "BalabanJ",
            "BertzCT",
            "Chi0",
            "Chi0n",
            "Chi0v",
            "Chi1",
            "Chi1n",
            "Chi1v",
            "Chi2n",
            "Chi2v",
            "Chi3n",
            "Chi3v",
            "Chi4n",
            "Chi4v",
            "HallKierAlpha",
            "Kappa1",
            "Kappa2",
            "Kappa3",
            "LabuteASA",
            "PEOE_VSA1",
            "PEOE_VSA10",
            "PEOE_VSA11",
            "PEOE_VSA12",
            "PEOE_VSA13",
            "PEOE_VSA14",
            "PEOE_VSA2",
            "PEOE_VSA3",
            "PEOE_VSA4",
            "PEOE_VSA5",
            "PEOE_VSA6",
            "PEOE_VSA7",
            "PEOE_VSA8",
            "PEOE_VSA9",
            "SMR_VSA1",
            "SMR_VSA10",
            "SMR_VSA2",
            "SMR_VSA3",
            "SMR_VSA4",
            "SMR_VSA5",
            "SMR_VSA6",
            "SMR_VSA7",
            "SMR_VSA8",
            "SMR_VSA9",
            "SlogP_VSA1",
            "SlogP_VSA10",
            "SlogP_VSA11",
            "SlogP_VSA12",
            "SlogP_VSA2",
            "SlogP_VSA3",
            "SlogP_VSA4",
            "SlogP_VSA5",
            "SlogP_VSA6",
            "SlogP_VSA7",
            "SlogP_VSA8",
            "SlogP_VSA9",
            "SPS",
            "TPSA",
            "EState_VSA1",
            "EState_VSA10",
            "EState_VSA11",
            "EState_VSA2",
            "EState_VSA3",
            "EState_VSA4",
            "EState_VSA5",
            "EState_VSA6",
            "EState_VSA7",
            "EState_VSA8",
            "EState_VSA9",
            "VSA_EState1",
            "VSA_EState10",
            "VSA_EState2",
            "VSA_EState3",
            "VSA_EState4",
            "VSA_EState5",
            "VSA_EState6",
            "VSA_EState7",
            "VSA_EState8",
            "VSA_EState9",
            "FractionCSP3",
            "HeavyAtomCount",
            "NHOHCount",
            "NOCount",
            "NumAliphaticCarbocycles",
            "NumAliphaticHeterocycles",
            "NumAliphaticRings",
            "NumAromaticCarbocycles",
            "NumAromaticHeterocycles",
            "NumAromaticRings",
            "NumHAcceptors",
            "NumHDonors",
            "NumHeteroatoms",
            "NumRotatableBonds",
            "NumSaturatedCarbocycles",
            "NumSaturatedHeterocycles",
            "NumSaturatedRings",
            "RingCount",
            "MolLogP",
            "MolMR",
            "fr_Al_COO",
            "fr_Al_OH",
            "fr_Al_OH_noTert",
            "fr_ArN",
            "fr_Ar_COO",
            "fr_Ar_N",
            "fr_Ar_NH",
            "fr_Ar_OH",
            "fr_COO",
            "fr_COO2",
            "fr_C_O",
            "fr_C_O_noCOO",
            "fr_C_S",
            "fr_HOCCN",
            "fr_Imine",
            "fr_NH0",
            "fr_NH1",
            "fr_NH2",
            "fr_N_O",
            "fr_Ndealkylation1",
            "fr_Ndealkylation2",
            "fr_Nhpyrrole",
            "fr_SH",
            "fr_aldehyde",
            "fr_alkyl_carbamate",
            "fr_alkyl_halide",
            "fr_allylic_oxid",
            "fr_amide",
            "fr_amidine",
            "fr_aniline",
            "fr_aryl_methyl",
            "fr_azide",
            "fr_azo",
            "fr_barbitur",
            "fr_benzene",
            "fr_benzodiazepine",
            "fr_bicyclic",
            "fr_diazo",
            "fr_dihydropyridine",
            "fr_epoxide",
            "fr_ester",
            "fr_ether",
            "fr_furan",
            "fr_guanido",
            "fr_halogen",
            "fr_hdrzine",
            "fr_hdrzone",
            "fr_imidazole",
            "fr_imide",
            "fr_isocyan",
            "fr_isothiocyan",
            "fr_ketone",
            "fr_ketone_Topliss",
            "fr_lactam",
            "fr_lactone",
            "fr_methoxy",
            "fr_morpholine",
            "fr_nitrile",
            "fr_nitro",
            "fr_nitro_arom",
            "fr_nitro_arom_nonortho",
            "fr_nitroso",
            "fr_oxazole",
            "fr_oxime",
            "fr_para_hydroxylation",
            "fr_phenol",
            "fr_phenol_noOrthoHbond",
            "fr_phos_acid",
            "fr_phos_ester",
            "fr_piperdine",
            "fr_piperzine",
            "fr_priamide",
            "fr_prisulfonamd",
            "fr_pyridine",
            "fr_quatN",
            "fr_sulfide",
            "fr_sulfonamd",
            "fr_sulfone",
            "fr_term_acetylene",
            "fr_tetrazole",
            "fr_thiazole",
            "fr_thiocyan",
            "fr_thiophene",
            "fr_unbrch_alkane",
            "fr_urea",
            "NumUnspecifiedAtomStereoCenters",
            "NumBridgeheadAtoms",
            "NumAmideBonds",
            "NumHeterocycles",
            "NumAtomStereoCenters",
            "Phi",
            "NumSpiroAtoms",
        }
        self.assertSetEqual(expected_descriptors, set(DEFAULT_DESCRIPTORS))

    def test_descriptor_calculation(self) -> None:
        """Test if the calculation of RDKitPhysChem Descriptors works as expected.

        Compared to precalculated values.

        Returns
        -------
        None
        """
        expected_df = pd.read_csv(data_path, sep="\t")
        descriptor_names = expected_df.drop(columns=["smiles"]).columns.tolist()
        smi2mol = SmilesToMol()
        property_element = MolToRDKitPhysChem(
            standardizer=None, descriptor_list=descriptor_names
        )
        pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("property_element", property_element),
            ]
        )
        smiles = expected_df["smiles"].tolist()
        property_vector = expected_df[descriptor_names].values

        output = pipeline.fit_transform(smiles)
        self.assertTrue(np.allclose(output, property_vector))  # add assertion here

    def test_descriptor_normalization(self) -> None:
        """Test if the normalization of RDKitPhysChem Descriptors works as expected.

        Returns
        -------
        None
        """
        smi2mol = SmilesToMol()
        property_element = MolToRDKitPhysChem(standardizer=StandardScaler())
        pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("property_element", property_element),
            ]
        )
        # pylint: disable=duplicate-code  # test case molecules are allowed to be duplicated
        smiles = [
            "CC",
            "CCC",
            "CCCO",
            "CCNCO",
            "C(C)CCO",
            "CCO",
            "CCCN",
            "CCCC",
            "CCOC",
            "COO",
        ]
        output = pipeline.fit_transform(smiles)
        non_zero_descriptors = output[:, (np.abs(output).sum(axis=0) != 0)]
        self.assertTrue(
            np.allclose(non_zero_descriptors.mean(axis=0), 0.0)
        )  # add assertion here
        self.assertTrue(np.allclose(non_zero_descriptors.std(axis=0), 1.0))

    def test_optional_nan_value_handling(self) -> None:
        """Test the handling of partly failed descriptor calculations."""

        ok_smiles_list = [
            "CC",
            "C(C)CCO",
        ]
        bad_smiles_list = [
            "F[P-](F)(F)(F)(F)F.CCCC[N+]1=CC=CC=C1C",
        ]

        # test with return_with_errors=False
        property_element = MolToRDKitPhysChem(
            standardizer=None, return_with_errors=False
        )

        error_filter = ErrorFilter.from_element_list([property_element])
        error_replacer = FilterReinserter.from_error_filter(
            error_filter, fill_value=np.nan
        )

        # note that we need the error filter and replacer here. Otherwise, the pipeline would fail on any error
        # irrespective of the return_with_errors parameter
        pipeline = Pipeline(
            [
                ("smi2mol", SmilesToMol()),
                ("property_element", property_element),
                ("error_filter", error_filter),
                ("error_replacer", error_replacer),
            ]
        )

        output = pipeline.fit_transform(bad_smiles_list + ok_smiles_list)
        self.assertEqual(len(output), len(bad_smiles_list + ok_smiles_list))
        # check expect-to-fail rows are ALL nan values
        self.assertTrue(
            np.equal(np.isnan(output).all(axis=1), [True, False, False]).all()
        )
        # check expected-not-to-fail rows contain zero nan values
        self.assertTrue(
            np.equal(np.isnan(output).any(axis=1), [True, False, False]).all()
        )

        # test with return_with_errors=True
        property_element2 = MolToRDKitPhysChem(
            standardizer=None, return_with_errors=True
        )

        error_filter2 = ErrorFilter.from_element_list([property_element2])
        filter_reinserter = FilterReinserter.from_error_filter(
            error_filter2, fill_value=np.nan
        )
        pipeline2 = Pipeline(
            [
                ("smi2mol", SmilesToMol()),
                ("property_element", property_element2),
                ("error_filter", error_filter2),
                ("error_replacer", filter_reinserter),
            ]
        )

        output2 = pipeline2.fit_transform(bad_smiles_list + ok_smiles_list)
        self.assertEqual(len(output2), len(bad_smiles_list + ok_smiles_list))
        # check expect-to-fail rows are ALL nan values
        self.assertTrue(
            np.equal(np.isnan(output2).all(axis=1), [False, False, False]).all()
        )
        # check expected-not-to-fail rows contain zero nan values
        self.assertTrue(
            np.equal(np.isnan(output2).any(axis=1), [True, False, False]).all()
        )

    def test_unknown_descriptor_name(self) -> None:
        """Test the handling of unknown descriptor names."""

        self.assertRaises(
            ValueError,
            MolToRDKitPhysChem,
            **{"descriptor_list": ["__NotADescriptor11Name:)"]},
        )

    def test_exception_handling(self) -> None:
        """Test exception handling during descriptor calculation."""

        pipeline = Pipeline(
            [
                ("smi2mol", SmilesToMol()),
                (
                    "property_element",
                    MolToRDKitPhysChem(
                        standardizer=None, return_with_errors=True, log_exceptions=False
                    ),
                ),
            ]
        )

        # Without exception handling [HH] would raise a division-by-zero exception because it has 0 heavy atoms
        output = pipeline.fit_transform(["[HH]"])
        self.assertTrue(output.shape == (1, len(DEFAULT_DESCRIPTORS)))


if __name__ == "__main__":
    unittest.main()



"""Wrapper for Chemprop loss functions."""

from typing import Any

import torch
from chemprop.nn.loss import BCELoss as _BCELoss
from chemprop.nn.loss import BinaryDirichletLoss as _BinaryDirichletLoss
from chemprop.nn.loss import CrossEntropyLoss as _CrossEntropyLoss
from chemprop.nn.loss import EvidentialLoss as _EvidentialLoss
from chemprop.nn.loss import LossFunction as _LossFunction
from chemprop.nn.loss import MSELoss as _MSELoss
from chemprop.nn.loss import MulticlassDirichletLoss as _MulticlassDirichletLoss
from chemprop.nn.loss import MVELoss as _MVELoss
from chemprop.nn.loss import SIDLoss as _SIDLoss
from numpy.typing import ArrayLike


class LossFunctionParamMixin:
    """Mixin for loss functions to get and set parameters."""

    _original_task_weights: ArrayLike

    def __init__(self: _LossFunction, task_weights: ArrayLike) -> None:
        """Initialize the loss function.

        Parameters
        ----------
        task_weights : ArrayLike
            The weights for each task.

        """
        super().__init__(task_weights=task_weights)  # type: ignore
        self._original_task_weights = task_weights

    # pylint: disable=unused-argument
    def get_params(self: _LossFunction, deep: bool = True) -> dict[str, Any]:
        """Get the parameters of the loss function.

        Parameters
        ----------
        deep : bool, optional
            Not used, only present to match the sklearn API.

        Returns
        -------
        dict[str, Any]
            The parameters of the loss function.
        """
        return {"task_weights": self._original_task_weights}

    def set_params(self: _LossFunction, **params: Any) -> _LossFunction:
        """Set the parameters of the loss function.

        Parameters
        ----------
        **params : Any
            The parameters to set.

        Returns
        -------
        Self
            The loss function with the new parameters.
        """
        task_weights = params.pop("task_weights", None)
        if task_weights is not None:
            self._original_task_weights = task_weights
            state_dict = self.state_dict()
            state_dict["task_weights"] = torch.as_tensor(
                task_weights, dtype=torch.float
            ).view(1, -1)
            self.load_state_dict(state_dict)
        return self


class BCELoss(LossFunctionParamMixin, _BCELoss):
    """Binary cross-entropy loss function."""


class BinaryDirichletLoss(LossFunctionParamMixin, _BinaryDirichletLoss):
    """Binary Dirichlet loss function."""


class CrossEntropyLoss(LossFunctionParamMixin, _CrossEntropyLoss):
    """Cross-entropy loss function."""


class EvidentialLoss(LossFunctionParamMixin, _EvidentialLoss):
    """Evidential loss function."""


class MSELoss(LossFunctionParamMixin, _MSELoss):
    """Mean squared error loss function."""


class MulticlassDirichletLoss(LossFunctionParamMixin, _MulticlassDirichletLoss):
    """Multiclass Dirichlet loss function."""


class MVELoss(LossFunctionParamMixin, _MVELoss):
    """Mean value entropy loss function."""


class SIDLoss(LossFunctionParamMixin, _SIDLoss):
    """SID loss function."""



"""Initialize the module for abstract mol2mol elements."""

from molpipeline.abstract_pipeline_elements.mol2mol.filter import (
    BaseKeepMatchesFilter,
    BasePatternsFilter,
)

__all__ = ["BasePatternsFilter", "BaseKeepMatchesFilter"]



"""This module tests the precomputed kernel estimators."""

from __future__ import annotations

import unittest

import numpy as np
import numpy.typing as npt
from rdkit import Chem
from rdkit.Chem import rdFingerprintGenerator
from rdkit.DataStructs import BulkTanimotoSimilarity
from scipy import sparse
from sklearn.neighbors import KNeighborsClassifier

from molpipeline import ErrorFilter, FilterReinserter, Pipeline, PostPredictionWrapper
from molpipeline.any2mol import SmilesToMol
from molpipeline.estimators import TanimotoToTraining
from molpipeline.mol2any import MolToMorganFP
from molpipeline.utils.kernel import tanimoto_similarity_sparse

COMPOUND_LIST = [
    "CCC",
    "CCOCC",
    "CCNCC",
    "CCO",
    "CCSC",
    "c1ccccc1",
    "c1c[nH]cc1",
]

IS_AROMATIC = [
    False,
    False,
    False,
    False,
    False,
    True,
    True,
]


def _generate_morgan_fingerprints(compound_list: list[str]) -> sparse.csr_matrix:
    """Generate the Morgan fingerprints.

    Parameters
    ----------
    compound_list: list[str]
        List of SMILES strings.

    Returns
    -------
    sparse.csr_matrix
        Morgan fingerprints.
    """
    morgan_pipeline = Pipeline(
        [
            ("smi2mol", SmilesToMol()),
            ("mol2fp", MolToMorganFP()),
        ]
    )
    fingerprint = morgan_pipeline.fit_transform(compound_list)
    return fingerprint


def _calculate_rdkit_self_similarity(
    compound_list: list[str],
) -> npt.NDArray[np.float64]:
    """Calculate the self similarity using RDKit.

    Parameters
    ----------
    compound_list: list[str]
        List of SMILES strings.

    Returns
    -------
    npt.NDArray[np.float64]
        Self similarity.
    """
    fp_list = []
    morgan_generator = rdFingerprintGenerator.GetMorganGenerator(radius=2)
    for smi in compound_list:
        mol = Chem.MolFromSmiles(smi)
        fp_list.append(morgan_generator.GetFingerprint(mol))
    sim = []
    for fp1 in fp_list:
        sim.append(BulkTanimotoSimilarity(fp1, fp_list))
    return np.array(sim)


class TestTanimotoSimilarityToTraining(unittest.TestCase):
    """Test for the TanimotoSimilarityToTraining."""

    def test_fit_transform(self) -> None:
        """Test if similarity calculation works for fit_transform."""
        # Reference: Only the fingerprint calculation
        fingerprint = _generate_morgan_fingerprints(COMPOUND_LIST)
        self_similarity = tanimoto_similarity_sparse(fingerprint, fingerprint)

        # Setup the full pipeline
        full_pipeline = Pipeline(
            [
                ("smi2mol", SmilesToMol()),
                ("mol2fp", MolToMorganFP()),
                ("precompute_tanimoto", TanimotoToTraining()),
            ]
        )

        # Test fit_transform
        pipeline_sim = full_pipeline.fit_transform(COMPOUND_LIST)
        self.assertTrue(np.allclose(pipeline_sim, self_similarity))

        # Test transform
        reversed_pipeline_sim = full_pipeline.transform(COMPOUND_LIST[::-1])
        reversed_self_similarity = tanimoto_similarity_sparse(
            fingerprint[::-1], fingerprint
        )
        self.assertTrue(np.allclose(reversed_pipeline_sim, reversed_self_similarity))

    def test_fit_and_transform(self) -> None:
        """Test if the similarity calculation works for fit and transform separately."""
        # Reference: Only the fingerprint calculation
        fingerprint = _generate_morgan_fingerprints(COMPOUND_LIST)
        self_similarity = tanimoto_similarity_sparse(fingerprint, fingerprint)

        # Setup the full pipeline
        full_pipeline = Pipeline(
            [
                ("smi2mol", SmilesToMol()),
                ("mol2fp", MolToMorganFP()),
                ("precompute_tanimoto", TanimotoToTraining()),
            ]
        )
        full_pipeline.fit(COMPOUND_LIST)
        pipeline_sim = full_pipeline.transform(COMPOUND_LIST)
        self.assertTrue(np.allclose(pipeline_sim, self_similarity))

    def test_fit_transform_rdkit(self) -> None:
        """Test if the similarity calculation matches the RDKit implementation for fit_transform."""
        # Reference: RDKit implementation
        self_similarity = _calculate_rdkit_self_similarity(COMPOUND_LIST)

        # Setup the full pipeline
        full_pipeline = Pipeline(
            [
                ("smi2mol", SmilesToMol()),
                ("mol2fp", MolToMorganFP()),
                ("precompute_tanimoto", TanimotoToTraining()),
            ]
        )
        pipeline_sim = full_pipeline.fit_transform(COMPOUND_LIST)
        self.assertTrue(np.allclose(pipeline_sim, self_similarity))

    def test_fit_and_transform_rdkit(self) -> None:
        """Test if the similarity calculation matches the RDKit implementation for fit and transform separately."""
        self_similarity = _calculate_rdkit_self_similarity(COMPOUND_LIST)

        # Setup the full pipeline
        full_pipeline = Pipeline(
            [
                ("smi2mol", SmilesToMol()),
                ("mol2fp", MolToMorganFP()),
                ("precompute_tanimoto", TanimotoToTraining()),
            ]
        )
        full_pipeline.fit(COMPOUND_LIST)
        pipeline_sim = full_pipeline.transform(COMPOUND_LIST)
        self.assertTrue(np.allclose(pipeline_sim, self_similarity))

    def test_nearest_neighbor_pipeline(self) -> None:
        """Test if the nearest neighbor pipeline works."""
        # Setup the full pipeline
        full_pipeline = Pipeline(
            [
                ("smi2mol", SmilesToMol()),
                ("mol2fp", MolToMorganFP()),
                ("precompute_tanimoto", TanimotoToTraining()),
                ("nearest_neighbor", KNeighborsClassifier(n_neighbors=1)),
            ]
        )
        full_pipeline.fit(COMPOUND_LIST, IS_AROMATIC)
        prediction = full_pipeline.predict(COMPOUND_LIST)
        self.assertTrue(np.allclose(prediction, IS_AROMATIC))

    def test_error_handling(self) -> None:
        """Test if the error handling works."""
        # Setup the full pipeline
        full_pipeline = Pipeline(
            [
                ("smi2mol", SmilesToMol()),
                ("mol2fp", MolToMorganFP()),
                ("precompute_tanimoto", TanimotoToTraining()),
                ("nearest_neighbor", KNeighborsClassifier(n_neighbors=1)),
            ]
        )
        full_pipeline.fit(COMPOUND_LIST, IS_AROMATIC)
        prediction = full_pipeline.predict(COMPOUND_LIST)
        self.assertTrue(np.allclose(prediction, IS_AROMATIC))

        error_filter = ErrorFilter(filter_everything=True)
        error_replacer = FilterReinserter.from_error_filter(
            error_filter, fill_value=np.nan
        )
        # Test if the error handling works
        full_pipeline = Pipeline(
            [
                ("smi2mol", SmilesToMol()),
                ("error_filter", error_filter),
                ("mol2fp", MolToMorganFP()),
                ("precompute_tanimoto", TanimotoToTraining()),
                ("nearest_neighbor", KNeighborsClassifier(n_neighbors=1)),
                ("error_replacer", PostPredictionWrapper(error_replacer)),
            ]
        )
        full_pipeline.fit(COMPOUND_LIST + ["C#C#C"], IS_AROMATIC + [False])
        prediction = full_pipeline.predict(COMPOUND_LIST + ["C#C#C"]).tolist()
        self.assertListEqual(prediction[:-1], IS_AROMATIC)
        self.assertTrue(np.isnan(prediction[-1]))

        single_prediction = full_pipeline.predict(["C#C#C"])
        self.assertTrue(np.isnan(single_prediction))

        none_prediction = full_pipeline.predict([])
        self.assertListEqual(none_prediction, [])

    def test_fit_transform_distance(self) -> None:
        """Test if distance calculation works for fit_transform."""
        # Reference: Only the fingerprint calculation
        fingerprint = _generate_morgan_fingerprints(COMPOUND_LIST)
        self_distance = 1 - tanimoto_similarity_sparse(fingerprint, fingerprint)

        # Setup the full pipeline
        full_pipeline = Pipeline(
            [
                ("smi2mol", SmilesToMol()),
                ("mol2fp", MolToMorganFP()),
                ("precompute_tanimoto", TanimotoToTraining(distance=True)),
            ]
        )

        # Test fit_transform
        pipeline_distance = full_pipeline.fit_transform(COMPOUND_LIST)
        self.assertTrue(np.allclose(pipeline_distance, self_distance))

        # Test transform
        reversed_pipeline_distance = full_pipeline.transform(COMPOUND_LIST[::-1])
        reversed_self_distance = 1 - tanimoto_similarity_sparse(
            fingerprint[::-1], fingerprint
        )
        self.assertTrue(np.allclose(reversed_pipeline_distance, reversed_self_distance))

    def test_fit_and_transform_distance(self) -> None:
        """Test if the distance calculation works for fit and transform separately."""
        # Reference: Only the fingerprint calculation
        fingerprint = _generate_morgan_fingerprints(COMPOUND_LIST)
        self_distance = 1 - tanimoto_similarity_sparse(fingerprint, fingerprint)

        # Setup the full pipeline
        full_pipeline = Pipeline(
            [
                ("smi2mol", SmilesToMol()),
                ("mol2fp", MolToMorganFP()),
                ("precompute_tanimoto", TanimotoToTraining(distance=True)),
            ]
        )
        full_pipeline.fit(COMPOUND_LIST)
        pipeline_distance = full_pipeline.transform(COMPOUND_LIST)
        self.assertTrue(np.allclose(pipeline_distance, self_distance))

    def test_fit_transform_rdkit_distance(self) -> None:
        """Test if the distance calculation matches the RDKit implementation for fit_transform."""
        # Reference: RDKit implementation
        self_distance = 1.0 - _calculate_rdkit_self_similarity(COMPOUND_LIST)

        # Setup the full pipeline
        full_pipeline = Pipeline(
            [
                ("smi2mol", SmilesToMol()),
                ("mol2fp", MolToMorganFP()),
                ("precompute_tanimoto", TanimotoToTraining(distance=True)),
            ]
        )
        pipeline_distance = full_pipeline.fit_transform(COMPOUND_LIST)
        self.assertTrue(np.allclose(pipeline_distance, self_distance))

    def test_fit_and_transform_rdkit_distance(self) -> None:
        """Test if the distance calculation matches the RDKit implementation for fit and transform separately."""
        # Reference: RDKit implementation
        self_distance = 1.0 - _calculate_rdkit_self_similarity(COMPOUND_LIST)

        # Setup the full pipeline
        full_pipeline = Pipeline(
            [
                ("smi2mol", SmilesToMol()),
                ("mol2fp", MolToMorganFP()),
                ("precompute_tanimoto", TanimotoToTraining(distance=True)),
            ]
        )
        full_pipeline.fit(COMPOUND_LIST)
        pipeline_distance = full_pipeline.transform(COMPOUND_LIST)
        self.assertTrue(np.allclose(pipeline_distance, self_distance))



"""Classes and functions for detecting and handling None values."""

from __future__ import annotations

from typing import Any, Generic, Iterable, Optional, Sequence, TypeVar

try:
    from typing import Self  # type: ignore[attr-defined]
except ImportError:
    from typing_extensions import Self

import numpy as np
import numpy.typing as npt

from molpipeline.abstract_pipeline_elements.core import (
    ABCPipelineElement,
    InvalidInstance,
    RemovedInstance,
    TransformingPipelineElement,
)
from molpipeline.utils.molpipeline_types import AnyVarSeq, TypeFixedVarSeq

__all__ = ["FilterReinserter", "ErrorFilter", "_MultipleErrorFilter"]

_T = TypeVar("_T")
_S = TypeVar("_S")


# pylint: disable=R0903
class ErrorFilter(ABCPipelineElement):
    """Collects None values and can fill Dummy values to matrices where None values were removed."""

    element_ids: set[str]
    error_indices: list[int]
    filter_everything: bool
    n_total: int

    def __init__(
        self,
        element_ids: set[str] | None = None,
        filter_everything: bool = True,
        name: str = "ErrorFilter",
        n_jobs: int = 1,
        uuid: Optional[str] = None,
    ) -> None:
        """Initialize NoneCollector.

        Parameters
        ----------
        element_ids: list[str]
            List of Pipeline Elements for which InvalidInstances can be removed.
        filter_everything: bool, optional (default: True)
            If True, element_ids are ignored and all InvalidInstances are removed.
        name: str, optional (default: "ErrorFilter")
            Name of the pipeline element.
        n_jobs: int, optional (default: 1)
            Number of parallel jobs to use.
        uuid: str, optional (default: None)
            UUID of the pipeline element.

        Returns
        -------
        None
        """
        super().__init__(name=name, n_jobs=n_jobs, uuid=uuid)
        self.error_indices = []
        if element_ids is None:
            if not filter_everything:
                raise ValueError(
                    "If element_ids is None, filter_everything must be True"
                )
            element_ids = set()
        if not isinstance(element_ids, set):
            element_ids = set(element_ids)
        self.element_ids = element_ids
        self.filter_everything = filter_everything
        self.n_total = 0
        self._requires_fitting = True

    @classmethod
    def from_element_list(
        cls,
        element_list: Iterable[TransformingPipelineElement],
        name: str = "ErrorFilter",
        n_jobs: int = 1,
        uuid: Optional[str] = None,
    ) -> Self:
        """
        Initialize NoneCollector from a list of Pipeline Elements.

        Parameters
        ----------
        element_list: Iterable[TransformingPipelineElement]
            List of Pipeline Elements for which None can be removed.
        name: str, optional (default: "ErrorFilter")
            Name of the pipeline element.
        n_jobs: int, optional (default: 1)
            Number of parallel jobs to use.
        uuid: str, optional (default: None)
            UUID of the pipeline element.

        Returns
        -------
        ErrorFilter
            Constructed ErrorFilter object.
        """
        element_ids = {element.uuid for element in element_list}
        return cls(
            element_ids, filter_everything=False, name=name, n_jobs=n_jobs, uuid=uuid
        )

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Get parameters for this element.

        Parameters
        ----------
        deep: bool
            If True, will return a deep copy of parameters for this estimator.

        Returns
        -------
        dict[str, Any]
            Parameter names mapped to their values.
        """
        params = super().get_params(deep=deep)
        params["filter_everything"] = self.filter_everything
        if deep:
            params["element_ids"] = {str(element_id) for element_id in self.element_ids}
        else:
            params["element_ids"] = self.element_ids
        return params

    def set_params(self, **parameters: Any) -> Self:
        """Set parameters for this element.

        Parameters
        ----------
        parameters: Any
            Dict of arameters to set.

        Returns
        -------
        Self
            Self with updated parameters.
        """
        param_copy = dict(parameters)
        if "element_ids" in param_copy:
            element_ids = param_copy.pop("element_ids")
            if not isinstance(element_ids, set):
                raise TypeError(f"Unexpected Type: {type(element_ids)}")
            self.element_ids = element_ids
        if "filter_everything" in param_copy:
            self.filter_everything = bool(param_copy.pop("filter_everything"))
        super().set_params(**param_copy)
        return self

    def check_removal(self, value: Any) -> bool:
        """Check if value should be removed.

        Parameters
        ----------
        value: AnyType
            Value to be checked.

        Returns
        -------
        bool
            True if value should be removed.
        """
        if not isinstance(value, InvalidInstance):
            return False
        if self.filter_everything or value.element_id in self.element_ids:
            return True
        return False

    def fit(
        self, values: AnyVarSeq, labels: Any = None
    ) -> Self:  # pylint: disable=unused-argument
        """Fit to input values.

        Only for compatibility with sklearn Pipelines.

        Parameters
        ----------
        values: AnyVarSeq
            Values used for fitting. (Not really used)
        labels: Any
            Label used for fitting. (Not really used)

        Returns
        -------
        Self
            Fitted ErrorFilter.
        """
        return self

    def fit_transform(
        self, values: TypeFixedVarSeq, labels: Any = None
    ) -> TypeFixedVarSeq:
        """Transform values and return a list without the None values.

        So far fit does nothing and hence is only called for consitency.

        Parameters
        ----------
        values: TypeFixedVarSeq
            Iterable to which element is fitted and which is subsequently transformed.
        labels: Any
            Label used for fitting. (Not used, but required for compatibility with sklearn)

        Returns
        -------
        TypeFixedVarSeq
            Iterable where invalid instances were removed.
        """
        self.fit(values, labels)
        return self.transform(values)

    def co_transform(self, values: TypeFixedVarSeq) -> TypeFixedVarSeq:
        """Remove rows at positions which contained discarded values during transformation.

        This ensures that rows of this instance maintain a one to one correspondence with the rows of data seen during
        transformation.

        Parameters
        ----------
        values: TypeFixedVarSeq
            Values to be transformed.

        Returns
        -------
        TypeFixedVarSeq
            Input where rows are removed.
        """
        if self.n_total != len(values):
            raise ValueError("Length of values does not match length of values in fit")
        if isinstance(values, list):
            out_list = []
            for idx, value in enumerate(values):
                if idx not in self.error_indices:
                    out_list.append(value)
            return out_list  # type: ignore  # mypy does not support higher-kinded types
        if isinstance(values, np.ndarray):
            return np.delete(values, self.error_indices, axis=0)  # type: ignore
        raise TypeError(f"Unexpected Type: {type(values)}")

    def transform(self, values: TypeFixedVarSeq) -> TypeFixedVarSeq:
        """Transform values and return a list without the None values.

        IMPORTANT: Changes number of elements in the iterable.

        Parameters
        ----------
        values: TypeFixedVarSeq
            Iterable of which according invalid instances are removed.

        Returns
        -------
        TypeFixedVarSeq
            Iterable where invalid instances were removed.
        """
        self.n_total = len(values)
        self.error_indices = []
        for i, value in enumerate(values):
            if self.check_removal(value):
                self.error_indices.append(i)
        return self.co_transform(values)

    def transform_single(self, value: Any) -> Any:
        """Transform a single value.

        Parameters
        ----------
        value: Any
            Value to be transformed.

        Returns
        -------
        Any
            Transformed value.
        """
        return self.pretransform_single(value)

    def pretransform_single(self, value: Any) -> Any:
        """Transform a single value.

        Parameters
        ----------
        value: Any
            Value to be transformed.

        Returns
        -------
        Any
            Transformed value.
        """
        if self.check_removal(value):
            return RemovedInstance(
                filter_element_id=self.uuid, message=value.message  # type: ignore
            )
        return value


class _MultipleErrorFilter:
    """Combines multiple ErrorFilters into one object."""

    error_filter_list: list[ErrorFilter]

    def __init__(self, error_filter_list: list[ErrorFilter]) -> None:
        """Initialize NoneCollector.

        Parameters
        ----------
        error_filter_list: list[ErrorFilter]
            List of ErrorFilter objects.
        """
        self.error_filter_list = error_filter_list
        prior_remover_dict = {}
        for i, remover_element in enumerate(error_filter_list):
            prior_remover_dict[remover_element] = error_filter_list[:i]
        self.prior_remover_dict = prior_remover_dict

    def transform(self, values: TypeFixedVarSeq) -> TypeFixedVarSeq:
        """Transform values and return a list without the invalid values.

        Parameters
        ----------
        values: TypeFixedVarSeq
            Iterable to which element is fitted and which is subsequently transformed.

        Returns
        -------
        TypeFixedVarSeq
            Iterable where invalid instances were removed.
        """
        for error_filter in self.error_filter_list:
            values = error_filter.transform(values)
        return values

    def co_transform(self, values: TypeFixedVarSeq) -> TypeFixedVarSeq:
        """Remove rows at positions which contained discarded values during transformation.

        Parameters
        ----------
        values: TypeFixedVarSeq
            Iterable to which element is fitted and which is subsequently transformed.

        Returns
        -------
        TypeFixedVarSeq
            Iterable where rows are removed which were removed during the transformation.
        """
        for error_filter in self.error_filter_list:
            values = error_filter.co_transform(values)
        return values

    def fit_transform(self, values: TypeFixedVarSeq) -> TypeFixedVarSeq:
        """Transform values and return a list without the None values.

        Parameters
        ----------
        values: TypeFixedVarSeq
            Iterable to which element is fitted and which is subsequently transformed.

        Returns
        -------
        TypeFixedVarSeq
            Iterable where invalid instances were removed.
        """
        for error_filter in self.error_filter_list:
            values = error_filter.fit_transform(values)
        return values

    def register_removed(self, index: int, value: RemovedInstance) -> None:
        """Register an invalid instance.

        Parameters
        ----------
        index: int
            Index of the invalid instance.
        value: Any
            Value of the invalid instance.
        """
        if not isinstance(value, RemovedInstance):
            raise TypeError(f"Unexpected Type: {type(value)}")

        for error_filter in self.error_filter_list:
            if value.filter_element_id == error_filter.uuid:
                new_index = index
                for prior_remover in self.prior_remover_dict[error_filter]:
                    new_index -= len(prior_remover.error_indices)
                error_filter.error_indices.append(index)
                break
        else:
            raise ValueError(
                f"Invalid instance not captured by any ErrorFilter: {value.filter_element_id}"
            )

    def set_total(self, total: int) -> None:
        """Set the total number of instances.

        Parameters
        ----------
        total: int
            Total number of instances seen during transformation.

        Returns
        -------
        None
        """
        for error_filter in self.error_filter_list:
            error_filter.n_total = total
            total -= len(error_filter.error_indices)


class FilterReinserter(ABCPipelineElement, Generic[_T]):
    """Fill None values with a Dummy value."""

    fill_value: _T
    error_filter_id: str
    _error_filter: Optional[ErrorFilter]
    n_total: int

    def __init__(
        self,
        error_filter_id: str,
        fill_value: _T,
        name: str = "FilterReinserter",
        n_jobs: int = 1,
        uuid: Optional[str] = None,
    ) -> None:
        """Initialize FilterReinserter.

        Parameters
        ----------
        error_filter_id: str
            Id of the ErrorFilter to use for filling removed values.
        fill_value: Any
            Value which is used to fill removed values.
        name: str, optional (default: "FilterReinserter")
            Name of the pipeline element.
        n_jobs: int, optional (default: 1)
            Number of parallel jobs to use.
        uuid: str, optional (default: None)
            UUID of the pipeline element.
        """
        super().__init__(name=name, n_jobs=n_jobs, uuid=uuid)
        self.error_filter_id = error_filter_id
        self._error_filter = None
        self.fill_value = fill_value

    @classmethod
    def from_error_filter(
        cls,
        error_filter: ErrorFilter,
        fill_value: _T,
        name: str = "FilterReinserter",
        n_jobs: int = 1,
        uuid: Optional[str] = None,
    ) -> Self:
        """Initialize FilterReinserter from a ErrorFilter object.

        Parameters
        ----------
        error_filter: ErrorFilter
            ErrorFilter to use for filling removed values.
        fill_value: Any
            Value which is used to fill removed values.
        name: str, optional (default: "FilterReinserter")
            Name of the pipeline element.
        n_jobs: int, optional (default: 1)
            Number of parallel jobs to use.
        uuid: str, optional (default: None)
            UUID of the pipeline element.

        Returns
        -------
        FilterReinserter
            Constructed FilterReinserter object.
        """
        filler = cls(
            error_filter_id=error_filter.uuid,
            fill_value=fill_value,
            name=name,
            n_jobs=n_jobs,
            uuid=uuid,
        )
        filler.error_filter = error_filter
        return filler

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Get parameters for this element.

        Parameters
        ----------
        deep: bool
            If True, will return a deep copy of parameters for this estimator.

        Returns
        -------
        dict[str, Any]
            Parameter names mapped to their values.
        """
        params = super().get_params(deep=deep)
        if deep:
            params["error_filter_id"] = str(self.error_filter_id)
            if self.fill_value is not None:
                params["fill_value"] = type(self.fill_value)(self.fill_value)  # type: ignore
            else:
                params["fill_value"] = None
        else:
            params["error_filter_id"] = self.error_filter_id
            params["fill_value"] = self.fill_value
        return params

    def set_params(self, **parameters: Any) -> Self:
        """Set parameters for this element.

        Parameters
        ----------
        parameters: Any
            Parameter dict.

        Returns
        -------
        self
            The instance itself.
        """
        parameter_copy = dict(parameters)
        if "error_filter_id" in parameter_copy:
            self.error_filter_id = str(parameter_copy.pop("error_filter_id"))
        if "fill_value" in parameter_copy:
            self.fill_value = parameter_copy.pop("fill_value")
        super().set_params(**parameter_copy)
        return self

    @property
    def error_filter(self) -> ErrorFilter:
        """Get the ErrorFilter connected to this FilterReinserter."""
        if self._error_filter is None:
            raise ValueError("ErrorFilter not set")
        return self._error_filter

    @error_filter.setter
    def error_filter(self, error_filter: ErrorFilter) -> None:
        """Set the ErrorFilter.

        Parameters
        ----------
        error_filter: ErrorFilter
            ErrorFilter to set.
        """
        self._error_filter = error_filter

    def select_error_filter(self, error_filter_list: list[ErrorFilter]) -> Self:
        """Select the ErrorFilter from a list of ErrorFilters.

        Parameters
        ----------
        error_filter_list: list[ErrorFilter]
            List of ErrorFilters to select from.

        Returns
        -------
        Self
            FilterReinserter with updated ErrorFilter.
        """
        for error_filter in error_filter_list:
            if error_filter.uuid == self.error_filter_id:
                self.error_filter = error_filter
                break
        else:
            raise ValueError(f"ErrorFilter with id {self.error_filter_id} not found")
        return self

    # pylint: disable=unused-argument
    def fit(
        self,
        values: TypeFixedVarSeq,
        labels: Any = None,
        **params: Any,
    ) -> Self:
        """Fit to input values.

        Only for compatibility with sklearn Pipelines.

        Parameters
        ----------
        values: TypeFixedVarSeq
            Values used for fitting.
        labels: Any
            Label used for fitting. (Not used, but required for compatibility with sklearn)
        **params: Any
            Additional keyword arguments. (Not used)

        Returns
        -------
        Self
            Fitted FilterReinserter.
        """
        return self

    # pylint: disable=unused-argument
    def fit_transform(
        self,
        values: TypeFixedVarSeq,
        labels: Any = None,
        **params: Any,
    ) -> TypeFixedVarSeq:
        """Transform values and return a list without the Invalid values.

        So far fit does nothing and hence is only called for consitency.

        Parameters
        ----------
        values: TypeFixedVarSeq
            Iterable to which element is fitted and which is subsequently transformed.
        labels: Any
            Label used for fitting. (Not used, but required for compatibility with sklearn)
        **params: Any
            Additional keyword arguments. (Not used)

        Returns
        -------
        TypeFixedVarSeq
            Iterable where invalid instances were removed.
        """
        self.fit(values)
        return self.transform(values)

    def transform_single(self, value: Any) -> Any:
        """Transform a single value.

        Parameters
        ----------
        value: Any
            Value to be transformed.

        Returns
        -------
        Any
            Transformed value.
        """
        return self.pretransform_single(value)

    def pretransform_single(self, value: Any) -> Any:
        """Transform a single value.

        Parameters
        ----------
        value: Any
            Value to be transformed.

        Returns
        -------
        Any
            Transformed value.
        """
        if (
            isinstance(value, RemovedInstance)
            and value.filter_element_id == self.error_filter.uuid
        ):
            return self.fill_value
        return value

    def transform(
        self, values: TypeFixedVarSeq, **params: Any  # pylint: disable=unused-argument
    ) -> TypeFixedVarSeq:
        """Transform iterable of values by removing invalid instances.

        IMPORTANT: Changes number of elements in the iterable.

        Parameters
        ----------
        values: TypeFixedVarSeq
            Iterable of which according invalid instances are removed.
        **params: Any
            Additional keyword arguments.

        Returns
        -------
        TypeFixedVarSeq
            Iterable where invalid instances were removed.
        """
        if len(values) != self.error_filter.n_total - len(
            self.error_filter.error_indices
        ):
            raise ValueError(
                f"Length of values does not match length of values in fit. "
                f"Expected: {self.error_filter.n_total - len(self.error_filter.error_indices)}  - Received :{len(values)}"
            )
        return self.fill_with_dummy(values)

    def _fill_list(self, list_to_fill: Sequence[_S]) -> Sequence[_S | _T]:
        """Fill a list with dummy values.

        Parameters
        ----------
        list_to_fill: list[Number]
            List to fill with dummy values.

        Returns
        -------
        list[Number]
            List where dummy values were inserted to replace instances which could not be processed.
        """
        filled_list: list[_S | _T] = []
        next_value_pos = 0
        for index in range(len(list_to_fill) + len(self.error_filter.error_indices)):
            if index in self.error_filter.error_indices:
                filled_list.append(self.fill_value)
            else:
                filled_list.append(list_to_fill[next_value_pos])
                next_value_pos += 1
        if len(list_to_fill) != next_value_pos:
            raise AssertionError()
        return filled_list

    def _fill_numpy_arr(self, value_array: npt.NDArray[Any]) -> npt.NDArray[Any]:
        """Fill a numpy array with dummy values.

        Parameters
        ----------
        value_array: npt.NDArray[Any]
            Numpy array to fill with dummy values.

        Returns
        -------
        npt.NDArray[Any]
            Numpy array where dummy values were inserted to replace instances which could not be processed.
        """
        fill_value = self.fill_value
        output_shape = list(value_array.shape)
        output_shape[0] += len(self.error_filter.error_indices)
        has_value_indices = np.ones(output_shape[0], dtype=bool)
        has_value_indices[self.error_filter.error_indices] = False

        try:
            dtype = np.common_type(value_array, np.array([self.fill_value]))
        except TypeError:
            dtype = np.object_

        output_matrix: npt.NDArray[Any]
        output_matrix = np.full(output_shape, fill_value, dtype=dtype)
        output_matrix[has_value_indices, ...] = value_array
        return output_matrix

    def fill_with_dummy(
        self,
        value_container: AnyVarSeq,
    ) -> AnyVarSeq:
        """Insert dummy values at the positions in the value container.

        Parameters
        ----------
        value_container: TypeFixedVarSeq
            Iterable to fill with dummy values.

        Returns
        -------
        AnyVarSeq
            Iterable where dummy values were inserted to replace molecules which could not be processed.
        """
        if isinstance(value_container, list):
            return self._fill_list(value_container)  # type: ignore
        if isinstance(value_container, np.ndarray):
            return self._fill_numpy_arr(value_container)  # type: ignore
        raise TypeError(f"Unexpected Type: {type(value_container)}")



"""Unittests for testing conversion of molecules to binary string."""

import unittest

from rdkit import Chem, rdBase

from molpipeline import Pipeline
from molpipeline.any2mol import SmilesToMol
from molpipeline.mol2any import MolToBinary

# pylint: disable=duplicate-code  # test case molecules are allowed to be duplicated
SMILES_ANTIMONY = "[SbH6+3]"
SMILES_BENZENE = "c1ccccc1"
SMILES_CHLOROBENZENE = "Clc1ccccc1"
SMILES_CL_BR = "NC(Cl)(Br)C(=O)O"
SMILES_METAL_AU = "OC[C@H]1OC(S[Au])[C@H](O)[C@@H](O)[C@@H]1O"

MOL_ANTIMONY = Chem.MolFromSmiles(SMILES_ANTIMONY)
MOL_BENZENE = Chem.MolFromSmiles(SMILES_BENZENE)
MOL_CHLOROBENZENE = Chem.MolFromSmiles(SMILES_CHLOROBENZENE)
MOL_CL_BR = Chem.MolFromSmiles(SMILES_CL_BR)
MOL_METAL_AU = Chem.MolFromSmiles(SMILES_METAL_AU)


class TestMol2Binary(unittest.TestCase):
    """Test case for testing conversion of molecules to binary string representation."""

    def test_mol_to_binary(self) -> None:
        """Test if smiles converted correctly to binary string."""

        test_smiles = [
            SMILES_ANTIMONY,
            SMILES_BENZENE,
            SMILES_CHLOROBENZENE,
            SMILES_CL_BR,
            SMILES_METAL_AU,
        ]
        expected_mols = [
            MOL_ANTIMONY,
            MOL_BENZENE,
            MOL_CHLOROBENZENE,
            MOL_CL_BR,
            MOL_METAL_AU,
        ]

        pipeline = Pipeline(
            [
                ("Smiles2Mol", SmilesToMol()),
                ("Mol2Binary", MolToBinary()),
            ]
        )
        log_block = rdBase.BlockLogs()
        binary_mols = pipeline.fit_transform(test_smiles)
        self.assertEqual(len(test_smiles), len(binary_mols))
        actual_mols = [Chem.Mol(mol) for mol in binary_mols]
        self.assertTrue(
            all(
                Chem.MolToInchi(smiles_mol) == Chem.MolToInchi(original_mol)
                for smiles_mol, original_mol in zip(actual_mols, expected_mols)
            )
        )
        del log_block

    def test_mol_to_binary_invalid_input(self) -> None:
        """Test how invalid input is handled."""

        pipeline = Pipeline(
            [
                ("Mol2Binary", MolToBinary()),
            ]
        )

        # test empty molecule
        binary_mols = pipeline.fit_transform([Chem.MolFromSmiles("")])
        self.assertEqual(len(binary_mols), 1)
        self.assertEqual(Chem.MolToSmiles(Chem.Mol(binary_mols[0])), "")

        # test None as input
        self.assertRaises(AttributeError, pipeline.fit_transform, [None])



"""Pipeline elements for converting instances to bool."""

from typing import Any

from molpipeline.abstract_pipeline_elements.core import (
    InvalidInstance,
    MolToAnyPipelineElement,
)


class MolToBool(MolToAnyPipelineElement):
    """
    Element to generate a bool array from input.

    Valid molecules are passed as True, InvalidInstances are passed as False.
    """

    def pretransform_single(self, value: Any) -> bool:
        """Transform a value to a bool representation.

        Parameters
        ----------
        value: Any
            Value to be transformed to bool representation.

        Returns
        -------
        str
            Binary representation of molecule.
        """
        if isinstance(value, InvalidInstance):
            return False
        return True

    def transform_single(self, value: Any) -> Any:
        """Transform a single molecule to a bool representation.

        Valid molecules are passed as True, InvalidInstances are passed as False.
        RemovedMolecule objects are passed without change, as no transformations are applicable.

        Parameters
        ----------
        value: Any
            Current representation of the molecule. (Eg. SMILES, RDKit Mol, ...)

        Returns
        -------
        Any
            Bool representation of the molecule.
        """
        pre_value = self.pretransform_single(value)
        return self.finalize_single(pre_value)



"""Union find algorithm."""

from __future__ import annotations

import numpy as np
import numpy.typing as npt


class UnionFindNode:
    """Union find node.

    A UnionFindNode is a node in a union find data structure, also called disjoint-set data structure.
    It stores a collection of non-overlapping sets and provides operations for merging these sets.
    It can be used to determine connected components in a graph.
    """

    def __init__(self) -> None:
        """Initialize union find node."""
        # initially, each node is its own parent
        self.parent = self
        # connected component number for each node. Needed at the end when connected components are getting extracted.
        self.connected_component_number = -1

    def find(self) -> UnionFindNode:
        """Find the root node.

        Returns
        -------
        Self
            Root node.
        """
        if self.parent is not self:
            self.parent = self.parent.find()
        return self.parent

    def union(self, other: UnionFindNode) -> UnionFindNode:
        """Union two nodes.

        Parameters
        ----------
        other: Self
            Other node.

        Returns
        -------
        UnionFindNode
            Root node of set or self.
        """
        # get the root nodes of the connected components of both nodes.
        # Additionally, compress the paths to the root nodes by setting others' parent to the root node.
        if self is self.parent:
            # this node is a parent. Meaning the root of the connected component. Let's overwrite it.
            self.parent = other.find()
        elif self is other:
            # this node is the other node. They are identical and therefore already in the same set.
            return self
        else:
            # add other node to this node's parent (and in addition to this connected component)
            self.parent = self.parent.union(other)
        return self.parent

    @staticmethod
    def get_connected_components(
        union_find_nodes: list[UnionFindNode],
    ) -> tuple[int, npt.NDArray[np.int32]]:
        """Get connected components from a union find node list.

        Parameters
        ----------
        union_find_nodes: list[UnionFindNode]
            List of union find nodes.

        Returns
        -------
        tuple[int, np.ndarray[int]]
            Number of connected components and connected component labels.
        """
        # results
        connected_components_counter = 0
        connected_components_array = np.empty(len(union_find_nodes), dtype=np.int32)

        for i, node in enumerate(union_find_nodes):
            root_parent = node.find()
            if root_parent.connected_component_number == -1:
                # found root node of a connected component. Annotate it with a connected component number.
                root_parent.connected_component_number = connected_components_counter
                connected_components_counter += 1
            connected_components_array[i] = root_parent.connected_component_number
        return connected_components_counter, connected_components_array



"""Test SHAP's TreeExplainer wrapper."""

import unittest

import numpy as np
import pandas as pd
from rdkit import Chem, rdBase
from sklearn.base import BaseEstimator, is_classifier, is_regressor
from sklearn.ensemble import (
    GradientBoostingClassifier,
    GradientBoostingRegressor,
    RandomForestClassifier,
    RandomForestRegressor,
)
from sklearn.linear_model import LinearRegression, LogisticRegression
from sklearn.svm import SVC, SVR

from molpipeline import ErrorFilter, FilterReinserter, Pipeline, PostPredictionWrapper
from molpipeline.abstract_pipeline_elements.core import RDKitMol
from molpipeline.any2mol import SmilesToMol
from molpipeline.experimental.explainability import (
    SHAPFeatureAndAtomExplanation,
    SHAPFeatureExplanation,
    SHAPKernelExplainer,
    SHAPTreeExplainer,
)
from molpipeline.experimental.explainability.explanation import AtomExplanationMixin
from molpipeline.mol2any import (
    MolToConcatenatedVector,
    MolToMorganFP,
    MolToRDKitPhysChem,
)
from molpipeline.mol2mol import SaltRemover
from molpipeline.utils.subpipeline import SubpipelineExtractor
from tests.test_experimental.test_explainability.utils import (
    construct_kernel_shap_kwargs,
)

TEST_SMILES = ["CC", "CCO", "COC", "c1ccccc1(N)", "CCC(-O)O", "CCCN"]
CONTAINS_OX = [0, 1, 1, 0, 1, 0]

TEST_SMILES_WITH_BAD_SMILES = [
    "CC",
    "CCO",
    "COC",
    "MY_FIRST_BAD_SMILES",
    "c1ccccc1(N)",
    "CCC(-O)O",
    "CCCN",
    "BAD_SMILES_2",
]
CONTAINS_OX_BAD_SMILES = [0, 1, 1, 0, 0, 1, 0, 1]

_RANDOM_STATE = 67056


class TestSHAPExplainers(unittest.TestCase):
    """Test SHAP's Explainer wrappers."""

    def _test_valid_explanation(
        self,
        explanation: SHAPFeatureExplanation | SHAPFeatureAndAtomExplanation,
        estimator: BaseEstimator,
        molecule_reader_subpipeline: Pipeline,
        nof_features: int,
        test_smiles: str,
        explainer: SHAPKernelExplainer | SHAPTreeExplainer,
    ) -> None:
        """Helper method to test if the explanation is valid and has the correct shape and content.

        Parameters
        ----------
        explanation : Explanation
            The explanation to be tested.
        estimator : BaseEstimator
            The estimator used in the pipeline.
        molecule_reader_subpipeline : Pipeline
            The subpipeline that extracts the molecule from the input data.
        nof_features : int
            The number of features in the feature vector.
        test_smiles : str
            The SMILES string of the molecule.
        explainer : SHAPKernelExplainer | SHAPTreeExplainer
            The explainer used to generate the explanation.
        """
        self.assertTrue(explanation.is_valid())

        self.assertIsInstance(explanation.feature_vector, np.ndarray)
        self.assertEqual(
            (nof_features,), explanation.feature_vector.shape  # type: ignore[union-attr]
        )

        # feature names should be a list of not empty strings
        self.assertTrue(
            all(
                isinstance(name, str) and len(name) > 0
                for name in explanation.feature_names  # type: ignore[union-attr]
            )
        )
        self.assertEqual(
            len(explanation.feature_names), explanation.feature_vector.shape[0]  # type: ignore
        )

        self.assertIsInstance(explanation.molecule, RDKitMol)
        self.assertEqual(
            Chem.MolToInchi(*molecule_reader_subpipeline.transform([test_smiles])),
            Chem.MolToInchi(explanation.molecule),
        )

        self.assertIsInstance(explanation.prediction, np.ndarray)
        self.assertIsInstance(explanation.feature_weights, np.ndarray)
        if is_regressor(estimator):
            self.assertTrue((1,), explanation.prediction.shape)  # type: ignore[union-attr]
            self.assertEqual(
                (nof_features,), explanation.feature_weights.shape  # type: ignore[union-attr]
            )
        elif is_classifier(estimator):
            self.assertTrue((2,), explanation.prediction.shape)  # type: ignore[union-attr]
            if isinstance(explainer, SHAPTreeExplainer) and isinstance(
                estimator, GradientBoostingClassifier
            ):
                # there is currently a bug in SHAP's TreeExplainer for GradientBoostingClassifier
                # https://github.com/shap/shap/issues/3177 returning only one feature weight
                # which is also based on log odds. This check is a workaround until the bug is fixed.
                self.assertEqual(
                    (nof_features,), explanation.feature_weights.shape  # type: ignore[union-attr]
                )
            elif isinstance(estimator, SVC):
                # SVC seems to be handled differently by SHAP. It returns only a one dimensional
                # feature array for binary classification.
                self.assertTrue(
                    (1,), explanation.prediction.shape  # type: ignore[union-attr]
                )
                self.assertEqual(
                    (nof_features,), explanation.feature_weights.shape  # type: ignore[union-attr]
                )
            else:
                # normal binary classification case
                self.assertEqual(
                    (nof_features, 2), explanation.feature_weights.shape  # type: ignore[union-attr]
                )
        else:
            raise ValueError("Error in unittest. Unsupported estimator.")

        if issubclass(type(explainer), AtomExplanationMixin):
            self.assertIsInstance(explanation.atom_weights, np.ndarray)  # type: ignore[union-attr]
            self.assertEqual(
                explanation.atom_weights.shape,  # type: ignore[union-attr]
                (explanation.molecule.GetNumAtoms(),),  # type: ignore[union-attr]
            )

    def test_explanations_fingerprint_pipeline(  # pylint: disable=too-many-locals
        self,
    ) -> None:
        """Test SHAP's TreeExplainer wrapper on MolPipeline's pipelines with fingerprints."""

        tree_estimators = [
            RandomForestClassifier(n_estimators=2, random_state=_RANDOM_STATE),
            RandomForestRegressor(n_estimators=2, random_state=_RANDOM_STATE),
            GradientBoostingClassifier(n_estimators=2, random_state=_RANDOM_STATE),
            GradientBoostingRegressor(n_estimators=2, random_state=_RANDOM_STATE),
        ]
        other_estimators = [
            SVC(kernel="rbf", probability=False, random_state=_RANDOM_STATE),
            SVR(kernel="linear"),
            LogisticRegression(random_state=_RANDOM_STATE),
            LinearRegression(),
        ]
        n_bits = 64

        explainer_types = [
            SHAPKernelExplainer,
            SHAPTreeExplainer,
        ]
        explainer_estimators = [tree_estimators + other_estimators, tree_estimators]

        for estimators, explainer_type in zip(
            explainer_estimators, explainer_types, strict=True
        ):

            # test explanations with different estimators
            for estimator in estimators:
                pipeline = Pipeline(
                    [
                        ("smi2mol", SmilesToMol()),
                        ("morgan", MolToMorganFP(radius=1, n_bits=n_bits)),
                        ("model", estimator),
                    ]
                )
                pipeline.fit(TEST_SMILES, CONTAINS_OX)

                # some explainers require additional kwargs
                explainer_kwargs = {}
                if explainer_type == SHAPKernelExplainer:
                    explainer_kwargs = construct_kernel_shap_kwargs(
                        pipeline, TEST_SMILES
                    )

                explainer = explainer_type(pipeline, **explainer_kwargs)
                explanations = explainer.explain(TEST_SMILES)
                self.assertEqual(len(explanations), len(TEST_SMILES))

                self.assertTrue(
                    issubclass(explainer.return_element_type_, AtomExplanationMixin)
                )

                # get the subpipeline that extracts the molecule from the input data
                mol_reader_subpipeline = SubpipelineExtractor(
                    pipeline
                ).get_molecule_reader_subpipeline()
                self.assertIsInstance(mol_reader_subpipeline, Pipeline)

                for i, explanation in enumerate(explanations):
                    self._test_valid_explanation(
                        explanation,
                        estimator,
                        mol_reader_subpipeline,  # type: ignore[arg-type]
                        n_bits,
                        TEST_SMILES[i],
                        explainer=explainer,  # type: ignore[arg-type]
                    )

    # pylint: disable=too-many-locals
    def test_explanations_pipeline_with_invalid_inputs(self) -> None:
        """Test SHAP's TreeExplainer wrapper with invalid inputs."""

        # estimators to test
        estimators = [
            RandomForestClassifier(n_estimators=2, random_state=_RANDOM_STATE),
            RandomForestRegressor(n_estimators=2, random_state=_RANDOM_STATE),
            GradientBoostingClassifier(n_estimators=2, random_state=_RANDOM_STATE),
            GradientBoostingRegressor(n_estimators=2, random_state=_RANDOM_STATE),
        ]

        # fill values considered invalid predictions
        invalid_fill_values = [None, np.nan, pd.NA]
        # fill values considered valid predictions (although outside the valid range)
        valid_fill_values = [0, 999]
        # fill values to test
        fill_values = invalid_fill_values + valid_fill_values

        n_bits = 64

        for estimator in estimators:
            for fill_value in fill_values:

                # pipeline with ErrorFilter
                error_filter1 = ErrorFilter()
                pipeline1 = Pipeline(
                    [
                        ("smi2mol", SmilesToMol()),
                        ("salt_remover", SaltRemover()),
                        ("error_filter", error_filter1),
                        ("morgan", MolToMorganFP(radius=1, n_bits=64)),
                        ("model", estimator),
                    ]
                )

                # pipeline with ErrorFilter and FilterReinserter
                error_filter2 = ErrorFilter()
                error_reinserter2 = PostPredictionWrapper(
                    FilterReinserter.from_error_filter(error_filter2, fill_value)
                )
                pipeline2 = Pipeline(
                    [
                        ("smi2mol", SmilesToMol()),
                        ("salt_remover", SaltRemover()),
                        ("error_filter", error_filter2),
                        ("morgan", MolToMorganFP(radius=1, n_bits=n_bits)),
                        ("model", estimator),
                        ("error_reinserter", error_reinserter2),
                    ]
                )

                for pipeline in [pipeline1, pipeline2]:

                    pipeline.fit(TEST_SMILES_WITH_BAD_SMILES, CONTAINS_OX_BAD_SMILES)

                    explainer = SHAPTreeExplainer(pipeline)
                    log_block = rdBase.BlockLogs()  # pylint: disable=unused-variable
                    explanations = explainer.explain(TEST_SMILES_WITH_BAD_SMILES)
                    del log_block
                    self.assertEqual(
                        len(explanations), len(TEST_SMILES_WITH_BAD_SMILES)
                    )

                    # get the subpipeline that extracts the molecule from the input data
                    mol_reader_subpipeline = SubpipelineExtractor(
                        pipeline
                    ).get_molecule_reader_subpipeline()
                    self.assertIsNotNone(mol_reader_subpipeline)

                    for i, explanation in enumerate(explanations):
                        if i in [3, 7]:
                            self.assertFalse(explanation.is_valid())
                            continue

                        self._test_valid_explanation(
                            explanation,
                            estimator,
                            mol_reader_subpipeline,  # type: ignore[arg-type]
                            n_bits,
                            TEST_SMILES_WITH_BAD_SMILES[i],
                            explainer=explainer,
                        )

    def test_explanations_pipeline_with_physchem(self) -> None:
        """Test SHAP's TreeExplainer wrapper on physchem feature vector."""

        estimators = [
            RandomForestClassifier(n_estimators=2, random_state=_RANDOM_STATE),
            RandomForestRegressor(n_estimators=2, random_state=_RANDOM_STATE),
            GradientBoostingClassifier(n_estimators=2, random_state=_RANDOM_STATE),
            GradientBoostingRegressor(n_estimators=2, random_state=_RANDOM_STATE),
        ]

        # test explanations with different estimators
        for estimator in estimators:
            pipeline = Pipeline(
                [
                    ("smi2mol", SmilesToMol()),
                    ("physchem", MolToRDKitPhysChem()),
                    ("model", estimator),
                ]
            )

            pipeline.fit(TEST_SMILES, CONTAINS_OX)

            explainer = SHAPTreeExplainer(pipeline)
            explanations = explainer.explain(TEST_SMILES)
            self.assertEqual(len(explanations), len(TEST_SMILES))

            # get the subpipeline that extracts the molecule from the input data
            mol_reader_subpipeline = SubpipelineExtractor(
                pipeline
            ).get_molecule_reader_subpipeline()
            self.assertIsNotNone(mol_reader_subpipeline)

            for i, explanation in enumerate(explanations):
                self._test_valid_explanation(
                    explanation,
                    estimator,
                    mol_reader_subpipeline,  # type: ignore[arg-type]
                    pipeline.named_steps["physchem"].n_features,
                    TEST_SMILES[i],
                    explainer=explainer,
                )

                self.assertEqual(
                    explanation.feature_names,
                    pipeline.named_steps["physchem"].feature_names,
                )

    def test_explanations_pipeline_with_concatenated_features(self) -> None:
        """Test SHAP's TreeExplainer wrapper on concatenated feature vector."""

        estimators = [
            RandomForestClassifier(n_estimators=2, random_state=_RANDOM_STATE),
            RandomForestRegressor(n_estimators=2, random_state=_RANDOM_STATE),
            GradientBoostingClassifier(n_estimators=2, random_state=_RANDOM_STATE),
            GradientBoostingRegressor(n_estimators=2, random_state=_RANDOM_STATE),
        ]

        n_bits = 64

        # test explanations with different estimators
        for estimator in estimators:
            pipeline = Pipeline(
                [
                    ("smi2mol", SmilesToMol()),
                    (
                        "features",
                        MolToConcatenatedVector(
                            [
                                (
                                    "RDKitPhysChem",
                                    MolToRDKitPhysChem(),
                                ),
                                (
                                    "MorganFP",
                                    MolToMorganFP(radius=1, n_bits=n_bits),
                                ),
                            ]
                        ),
                    ),
                    ("model", estimator),
                ]
            )

            pipeline.fit(TEST_SMILES, CONTAINS_OX)

            explainer = SHAPTreeExplainer(pipeline)
            explanations = explainer.explain(TEST_SMILES)
            self.assertEqual(len(explanations), len(TEST_SMILES))

            # get the subpipeline that extracts the molecule from the input data
            mol_reader_subpipeline = SubpipelineExtractor(
                pipeline
            ).get_molecule_reader_subpipeline()
            self.assertIsNotNone(mol_reader_subpipeline)

            for i, explanation in enumerate(explanations):
                self._test_valid_explanation(
                    explanation,
                    estimator,
                    mol_reader_subpipeline,  # type: ignore[arg-type]
                    pipeline.named_steps["features"].n_features,
                    TEST_SMILES[i],
                    explainer=explainer,
                )

                self.assertEqual(
                    explanation.feature_names,
                    pipeline.named_steps["features"].feature_names,
                )



"""Script that runs all notebooks once to check if they execute without error."""

import argparse
import subprocess
import sys
from pathlib import Path

# list of directories containing notebooks
NOTEBOOK_DIRS = ["notebooks"]

# list of prefix paths of notebooks to skip. We mainly skip because some notebooks run too long.
SKIP_NOTEBOOKS_PREFIXES = ["notebooks/advanced_01"]


def is_prefix(prefix: Path, path: Path) -> bool:
    """Check if the prefix is a prefix of the path.

    Parameters
    ----------
    prefix: Path
        Prefix path.
    path: Path
        Path to check if it has the prefix.

    Returns
    -------
    bool
        True if the prefix is a prefix of the path, False otherwise.
    """
    # compares the parents of the paths are equal and the name of the path starts with the prefix file name
    return all(
        parent1 == parent2
        for parent1, parent2 in zip(
            prefix.resolve().parents, path.resolve().parents, strict=True
        )
    ) and path.name.startswith(prefix.name)


def get_notebook_paths_from_dir(notebook_dir: Path) -> list[Path]:
    """Get all Jupyter notebook files in the directory.

    Parameters
    ----------
    notebook_dir: Path
        Path to the directory containing the Jupyter notebooks.

    Returns
    -------
    list[Path]
        List of paths to Jupyter notebooks.
    """
    # Find all Jupyter notebook files in the directory
    notebooks_paths = []
    for notebook_path in notebook_dir.rglob("*.ipynb"):

        if ".ipynb_checkpoints" in str(notebook_path.resolve()):
            # skip jetbrains checkpoints
            continue
        if notebook_path.name.endswith(".nbconvert.ipynb"):
            # skip converted notebooks
            continue

        notebooks_paths.append(notebook_path)
    return notebooks_paths


def filter_notebooks(
    notebooks_paths: list[Path], skip_notebooks_prefixes: list[Path]
) -> list[Path]:
    """Filter out notebooks that should be skipped.

    Parameters
    ----------
    notebooks_paths: list[Path]
        List of paths to Jupyter notebooks.
    skip_notebooks_prefixes: list[Path]
        List of prefixes of notebooks to skip.

    Returns
    -------
    list[Path]
        Filtered list of paths to Jupyter notebooks.
    """
    return list(
        filter(
            lambda notebook_path: all(
                not is_prefix(prefix, notebook_path)
                for prefix in skip_notebooks_prefixes
            ),
            notebooks_paths,
        )
    )


def run_notebooks(
    notebook_dir: Path, skip_notebooks_prefixes: list[Path], continue_on_error: bool
) -> None:
    """Run all Jupyter notebooks in the given directory and check if they execute without error.

    Parameters
    ----------
    notebook_dir: Path
        Path to the directory containing the Jupyter notebooks.
    skip_notebooks_prefixes: list[Path]
        List of prefixes of notebooks to skip.
    continue_on_error: bool
        If True continue running all notebooks even if an error occurs.
    """
    # collect all notebook files from disc
    notebooks_paths = get_notebook_paths_from_dir(notebook_dir)

    # filter out notebooks that should be skipped
    notebooks_paths = sorted(filter_notebooks(notebooks_paths, skip_notebooks_prefixes))

    nof_errors = 0
    # Loop through each notebook
    for notebooks_path in notebooks_paths:

        # Execute the notebook and capture the error code
        cmd = [
            "jupyter",
            "nbconvert",
            "--execute",
            str(notebooks_path),
            "--to",
            "notebook",
        ]
        with subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        ) as process:
            std_out, std_err = process.communicate()
            error_code = process.returncode

        # Check if the error code is not 0
        if error_code != 0:
            nof_errors += 1
            log_error_msg = (
                f"Error executing {notebooks_path}. Error code: {error_code}"
            )
            print(log_error_msg)
            print(std_out.decode("utf-8"))
            print(std_err.decode("utf-8"))

            if not continue_on_error:
                sys.exit(1)

    # Check if there were any errors
    if nof_errors == 0:
        print("All Jupyter notebooks executed successfully with error code 0.")
    else:
        print(f"Jupyter notebooks executed with {nof_errors} errors.")
        sys.exit(1)


def main() -> None:
    """Main function to run the Jupyter notebooks."""
    parser = argparse.ArgumentParser(
        description="Test if all Jupyter notebooks in a directory run through with error code 0"
    )
    parser.add_argument(
        "-c",
        "--continue-on-failure",
        dest="continue_on_error",
        action="store_true",
        help="Continue running all notebooks even if an error occurs",
    )
    parser.set_defaults(continue_on_error=False)
    args = parser.parse_args()

    skip_notebook_prefixes_paths = [Path(prefix) for prefix in SKIP_NOTEBOOKS_PREFIXES]

    for notebook_dir in NOTEBOOK_DIRS:
        run_notebooks(
            Path(notebook_dir), skip_notebook_prefixes_paths, args.continue_on_error
        )


if __name__ == "__main__":
    main()



"""Abstract classes for filters."""

import abc
from typing import Any, Literal, Mapping, Optional, Sequence, TypeAlias, Union

try:
    from typing import Self  # type: ignore[attr-defined]
except ImportError:
    from typing_extensions import Self

from molpipeline.abstract_pipeline_elements.core import (
    InvalidInstance,
    MolToMolPipelineElement,
    OptionalMol,
    RDKitMol,
)
from molpipeline.utils.molpipeline_types import (
    FloatCountRange,
    IntCountRange,
    IntOrIntCountRange,
)
from molpipeline.utils.value_conversions import count_value_to_tuple

# possible mode types for a KeepMatchesFilter:
# - "any" means one match is enough
# - "all" means all elements must be matched
FilterModeType: TypeAlias = Literal["any", "all"]


def _within_boundaries(
    lower_bound: Optional[float], upper_bound: Optional[float], property_value: float
) -> bool:
    """Check if a value is within the specified boundaries.

    Boundaries given as None are ignored.

    Parameters
    ----------
    lower_bound: Optional[float]
        Lower boundary.
    upper_bound: Optional[float]
        Upper boundary.
    property_value: float
        Property value to check.

    Returns
    -------
    bool
        True if the value is within the boundaries, else False.
    """
    if lower_bound is not None and property_value < lower_bound:
        return False
    if upper_bound is not None and property_value > upper_bound:
        return False
    return True


class BaseKeepMatchesFilter(MolToMolPipelineElement, abc.ABC):
    """Filter to keep or remove molecules based on patterns.

    Notes
    -----
    There are four possible scenarios:
        - mode = "any" & keep_matches = True: Needs to match at least one filter element.
        - mode = "any" & keep_matches = False: Must not match any filter element.
        - mode = "all" & keep_matches = True: Needs to match all filter elements.
        - mode = "all" & keep_matches = False: Must not match all filter elements.
    """

    keep_matches: bool
    mode: FilterModeType

    def __init__(
        self,
        filter_elements: Union[
            Mapping[Any, Union[FloatCountRange, IntCountRange, IntOrIntCountRange]],
            Sequence[Any],
        ],
        keep_matches: bool = True,
        mode: FilterModeType = "any",
        name: Optional[str] = None,
        n_jobs: int = 1,
        uuid: Optional[str] = None,
    ) -> None:
        """Initialize BasePatternsFilter.

        Parameters
        ----------
        filter_elements: Union[Mapping[Any, Union[FloatCountRange, IntCountRange, IntOrIntCountRange]], Sequence[Any]]
            List of filter elements. Typically can be a list of patterns or a dictionary with patterns as keys and
            an int for exact count or a tuple of minimum and maximum.
            NOTE: for each child class, the type of filter_elements must be specified by the filter_elements setter.
        keep_matches: bool, optional (default: True)
            If True, molecules containing the specified patterns are kept, else removed.
        mode: FilterModeType, optional (default: "any")
            If "any", at least one of the specified patterns must be present in the molecule.
            If "all", all of the specified patterns must be present in the molecule.
        name: Optional[str], optional (default: None)
            Name of the pipeline element.
        n_jobs: int, optional (default: 1)
            Number of parallel jobs to use.
        uuid: str, optional (default: None)
            Unique identifier of the pipeline element.
        """
        super().__init__(name=name, n_jobs=n_jobs, uuid=uuid)
        self.filter_elements = filter_elements  # type: ignore
        self.keep_matches = keep_matches
        self.mode = mode

    @property
    @abc.abstractmethod
    def filter_elements(
        self,
    ) -> Mapping[Any, FloatCountRange]:
        """Get filter elements as dict."""

    @filter_elements.setter
    @abc.abstractmethod
    def filter_elements(
        self,
        filter_elements: Union[Mapping[Any, FloatCountRange], Sequence[Any]],
    ) -> None:
        """Set filter elements as dict.

        Parameters
        ----------
        filter_elements: Union[Mapping[Any, FloatCountRange], Sequence[Any]]
            List of filter elements.
        """

    def set_params(self, **parameters: Any) -> Self:
        """Set parameters of BaseKeepMatchesFilter.

        Parameters
        ----------
        parameters: Any
            Parameters to set.

        Returns
        -------
        Self
            Self.
        """
        parameter_copy = dict(parameters)
        if "keep_matches" in parameter_copy:
            self.keep_matches = parameter_copy.pop("keep_matches")
        if "mode" in parameter_copy:
            self.mode = parameter_copy.pop("mode")
        if "filter_elements" in parameter_copy:
            self.filter_elements = parameter_copy.pop("filter_elements")
        super().set_params(**parameter_copy)
        return self

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Get parameters of PatternFilter.

        Parameters
        ----------
        deep: bool, optional (default: True)
            If True, return the parameters of all subobjects that are PipelineElements.

        Returns
        -------
        dict[str, Any]
            Parameters of BaseKeepMatchesFilter.
        """
        params = super().get_params(deep=deep)
        params["keep_matches"] = self.keep_matches
        params["mode"] = self.mode
        params["filter_elements"] = self.filter_elements
        return params

    def pretransform_single(  # pylint: disable=too-many-return-statements
        self, value: RDKitMol
    ) -> OptionalMol:
        """Invalidate or validate molecule based on specified filter.

        There are four possible scenarios:
        - mode = "any" & keep_matches = True: Needs to match at least one filter element.
        - mode = "any" & keep_matches = False: Must not match any filter element.
        - mode = "all" & keep_matches = True: Needs to match all filter elements.
        - mode = "all" & keep_matches = False: Must not match all filter elements.

        Parameters
        ----------
        value: RDKitMol
            Molecule to check.

        Returns
        -------
        OptionalMol
            Molecule that matches defined filter elements, else InvalidInstance.
        """
        for filter_element, (lower_limit, upper_limit) in self.filter_elements.items():
            property_value = self._calculate_single_element_value(filter_element, value)
            if _within_boundaries(lower_limit, upper_limit, property_value):
                # For "any" mode we can return early if a match is found
                if self.mode == "any":
                    if not self.keep_matches:
                        return InvalidInstance(
                            self.uuid,
                            f"Molecule contains forbidden filter element {filter_element}.",
                            self.name,
                        )
                    return value
            else:
                # For "all" mode we can return early if a match is not found
                if self.mode == "all":
                    if self.keep_matches:
                        return InvalidInstance(
                            self.uuid,
                            f"Molecule does not contain required filter element {filter_element}.",
                            self.name,
                        )
                    return value

        # If this point is reached, no or all patterns were found
        # If mode is "any", finishing the loop means no match was found
        if self.mode == "any":
            if self.keep_matches:
                return InvalidInstance(
                    self.uuid,
                    "Molecule does not match any of the required filter elements.",
                    self.name,
                )
            #  else: No match with forbidden filter elements was found, return original molecule
            return value

        if self.mode == "all":
            if not self.keep_matches:
                return InvalidInstance(
                    self.uuid,
                    "Molecule matches all forbidden filter elements.",
                    self.name,
                )
            #  else: All required filter elements were found, return original molecule
            return value

        raise ValueError(f"Invalid mode: {self.mode}")

    @abc.abstractmethod
    def _calculate_single_element_value(
        self, filter_element: Any, value: RDKitMol
    ) -> float:
        """Calculate the value of a single match.

        Parameters
        ----------
        filter_element: Any
            Match case to calculate.
        value: RDKitMol
            Molecule to calculate the match for.

        Returns
        -------
        float
            Value of the match.
        """


class BasePatternsFilter(BaseKeepMatchesFilter, abc.ABC):
    """Filter to keep or remove molecules based on patterns.

    Attributes
    ----------
    filter_elements: Union[Sequence[str], Mapping[str, IntOrIntCountRange]]
        List of patterns to allow in molecules.
        Alternatively, a dictionary can be passed with patterns as keys
        and an int for exact count or a tuple of minimum and maximum.
    [...]

    Notes
    -----
    There are four possible scenarios:
    - mode = "any" & keep_matches = True: Needs to match at least one filter element.
    - mode = "any" & keep_matches = False: Must not match any filter element.
    - mode = "all" & keep_matches = True: Needs to match all filter elements.
    - mode = "all" & keep_matches = False: Must not match all filter elements.
    """

    _filter_elements: Mapping[str, IntCountRange]

    @property
    def filter_elements(self) -> Mapping[str, IntCountRange]:
        """Get allowed filter elements (patterns) as dict."""
        return self._filter_elements

    @filter_elements.setter
    def filter_elements(
        self,
        patterns: Union[list[str], Mapping[str, IntOrIntCountRange]],
    ) -> None:
        """Set allowed filter elements (patterns) as dict.

        Parameters
        ----------
        patterns: Union[list[str], Mapping[str, IntOrIntCountRange]]
            List of patterns.
        """
        if isinstance(patterns, (list, set)):
            self._filter_elements = {pat: (1, None) for pat in patterns}
        else:
            self._filter_elements = {
                pat: count_value_to_tuple(count) for pat, count in patterns.items()
            }
        self.patterns_mol_dict = list(self._filter_elements.keys())  # type: ignore

    @property
    def patterns_mol_dict(self) -> Mapping[str, RDKitMol]:
        """Get patterns as dict with RDKitMol objects."""
        return self._patterns_mol_dict

    @patterns_mol_dict.setter
    def patterns_mol_dict(self, patterns: Sequence[str]) -> None:
        """Set patterns as dict with RDKitMol objects.

        Parameters
        ----------
        patterns: Sequence[str]
            List of patterns.
        """
        self._patterns_mol_dict = {pat: self._pattern_to_mol(pat) for pat in patterns}
        failed_patterns = [
            pat for pat, mol in self._patterns_mol_dict.items() if not mol
        ]
        if failed_patterns:
            raise ValueError("Invalid pattern(s): " + ", ".join(failed_patterns))

    @abc.abstractmethod
    def _pattern_to_mol(self, pattern: str) -> RDKitMol:
        """Convert pattern to Rdkitmol object.

        Parameters
        ----------
        pattern: str
            Pattern to convert.

        Returns
        -------
        RDKitMol
            RDKitMol object of the pattern.
        """

    def _calculate_single_element_value(
        self, filter_element: Any, value: RDKitMol
    ) -> int:
        """Calculate a single match count for a molecule.

        Parameters
        ----------
        filter_element: Any
            smarts to calculate match count for.
        value: RDKitMol
            Molecule to calculate smarts match count for.

        Returns
        -------
        int
            smarts match count value.
        """
        return len(value.GetSubstructMatches(self.patterns_mol_dict[filter_element]))



"""Test connected component clustering algorithm."""

from __future__ import annotations

import unittest
from typing import Any

import numpy as np
from scipy.sparse import csr_matrix

from molpipeline.estimators.algorithm.connected_component_clustering import (
    calc_chunk_size_from_memory_requirement,
    connected_components_iterative_algorithm,
)


class TestConnectedComponentClusteringAlgorithm(unittest.TestCase):
    """Test connected component clustering algorithm."""

    def test_connected_components_iterative(self) -> None:
        """Test the iterative connected component clustering algorithm."""

        # test data: all possible fingerprints of len 3 but duplicated
        fingerprint_matrix = csr_matrix(
            [
                [0, 0, 0],  # 0
                [1, 0, 0],  # 1
                [1, 1, 0],  # 2
                [1, 0, 1],  # 3
                [0, 1, 0],  # 4
                [0, 1, 1],  # 5
                [0, 0, 1],  # 6
                [1, 1, 1],  # 7
                [0, 0, 0],  # 8
                [1, 0, 0],  # 9
                [1, 1, 0],  # 10
                [1, 0, 1],  # 11
                [0, 1, 0],  # 12
                [0, 1, 1],  # 13
                [0, 0, 1],  # 14
                [1, 1, 1],  # 15
            ]
        )

        eps: float = 1e-10

        expected_clusterings: list[dict[str, Any]] = [
            {
                "threshold": 0.0,
                "expected_clustering": [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            },
            {
                "threshold": 0.0 + eps,
                "expected_clustering": [0, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1],
            },
            {
                "threshold": 1 / 3 - eps,
                "expected_clustering": [0, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1],
            },
            {
                "threshold": 1 / 3,
                "expected_clustering": [0, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1],
            },
            {
                "threshold": 1 / 3 + eps,
                "expected_clustering": [0, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1],
            },
            {
                "threshold": 0.5 - eps,
                "expected_clustering": [0, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1],
            },
            {
                "threshold": 0.5,
                "expected_clustering": [0, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1],
            },
            {
                "threshold": 0.5 + eps,
                "expected_clustering": [0, 1, 2, 2, 3, 2, 4, 2, 5, 1, 2, 2, 3, 2, 4, 2],
            },
            {
                "threshold": 2 / 3 - eps,
                "expected_clustering": [0, 1, 2, 2, 3, 2, 4, 2, 5, 1, 2, 2, 3, 2, 4, 2],
            },
            {
                "threshold": 2 / 3,
                "expected_clustering": [0, 1, 2, 2, 3, 2, 4, 2, 5, 1, 2, 2, 3, 2, 4, 2],
            },
            {
                "threshold": 2 / 3 + eps,
                "expected_clustering": [0, 1, 2, 3, 4, 5, 6, 7, 8, 1, 2, 3, 4, 5, 6, 7],
            },
            {
                "threshold": 1.0 - eps,
                "expected_clustering": [0, 1, 2, 3, 4, 5, 6, 7, 8, 1, 2, 3, 4, 5, 6, 7],
            },
            {
                "threshold": 1.0,
                "expected_clustering": [0, 1, 2, 3, 4, 5, 6, 7, 8, 1, 2, 3, 4, 5, 6, 7],
            },
        ]

        for test_case_dict in expected_clusterings:
            threshold = test_case_dict["threshold"]
            expected_clustering = test_case_dict["expected_clustering"]

            self.assertEqual(fingerprint_matrix.shape[0], len(expected_clustering))

            nof_cc, cc_labels = connected_components_iterative_algorithm(
                fingerprint_matrix, threshold
            )

            exp_nof_clusters = np.unique(expected_clustering).shape[0]
            self.assertEqual(exp_nof_clusters, nof_cc)

            self.assertTrue(np.equal(cc_labels, expected_clustering).all())

    def test_connected_components_iterative_chunking(self) -> None:
        """Test the chunking of the iterative connected component clustering algorithm."""
        # test standard case: two samples with a similarity of >=0.5 group into the same cluster

        fingerprint_matrix = csr_matrix([[1, 0, 1], [0, 1, 0], [1, 0, 1]])

        nof_cc, cc_labels = connected_components_iterative_algorithm(
            fingerprint_matrix, 0.5, 2
        )
        self.assertEqual(nof_cc, 2)
        self.assertTrue(np.equal(cc_labels, [0, 1, 0]).all())

        # test equals case: two samples with a similarity of 1 group into the same cluster
        nof_cc, cc_labels = connected_components_iterative_algorithm(
            fingerprint_matrix, 1.0, 2
        )
        self.assertEqual(nof_cc, 2)
        self.assertTrue(np.equal(cc_labels, [0, 1, 0]).all())

        # test chunk size 1
        nof_cc, cc_labels = connected_components_iterative_algorithm(
            fingerprint_matrix, 0.5, 1
        )
        self.assertEqual(nof_cc, 2)
        self.assertTrue(np.equal(cc_labels, [0, 1, 0]).all())

        # test chunk size 3
        nof_cc, cc_labels = connected_components_iterative_algorithm(
            fingerprint_matrix, 0.5, 3
        )
        self.assertEqual(nof_cc, 2)
        self.assertTrue(np.equal(cc_labels, [0, 1, 0]).all())

        # test chunk size 4
        nof_cc, cc_labels = connected_components_iterative_algorithm(
            fingerprint_matrix, 0.5, 4
        )
        self.assertEqual(nof_cc, 2)
        self.assertTrue(np.equal(cc_labels, [0, 1, 0]).all())

    def test_calc_chunk_size_from_memory_requirement_for_tanimoto_similarity_sparse(
        self,
    ) -> None:
        """Test the calculation of the chunk size from the memory requirement."""
        matrix = csr_matrix([[1, 0, 1], [0, 1, 0], [1, 0, 1]])
        self.assertEqual(matrix.todense().nbytes, 72)

        # test memory_cutoff less than memory of row
        self.assertEqual(
            calc_chunk_size_from_memory_requirement(
                matrix.shape[0],
                matrix.shape[1],
                matrix.dtype.itemsize,
                23 / (1 << 30),  # 23 byte in GB. A row is 24 byte.
            ),
            1,
        )
        self.assertEqual(
            calc_chunk_size_from_memory_requirement(
                matrix.shape[0],
                matrix.shape[1],
                matrix.dtype.itemsize,
                24 / (1 << 30),  # 24 byte in GB. Exactly a row.
            ),
            1,
        )
        self.assertEqual(
            calc_chunk_size_from_memory_requirement(
                matrix.shape[0],
                matrix.shape[1],
                matrix.dtype.itemsize,
                25 / (1 << 30),  # 25 byte in GB. Slightly more than one row.
            ),
            1,
        )
        self.assertEqual(
            calc_chunk_size_from_memory_requirement(
                matrix.shape[0],
                matrix.shape[1],
                matrix.dtype.itemsize,
                47 / (1 << 30),  # 47 byte in GB. Almost two rows.
            ),
            1,
        )
        self.assertEqual(
            calc_chunk_size_from_memory_requirement(
                matrix.shape[0],
                matrix.shape[1],
                matrix.dtype.itemsize,
                48 / (1 << 30),  # 48 byte in GB. Exactly two rows.
            ),
            2,
        )
        self.assertEqual(
            calc_chunk_size_from_memory_requirement(
                matrix.shape[0],
                matrix.shape[1],
                matrix.dtype.itemsize,
                49 / (1 << 30),  # 49 byte in GB. Slightly more than two rows.
            ),
            2,
        )
        self.assertEqual(
            calc_chunk_size_from_memory_requirement(
                matrix.shape[0],
                matrix.shape[1],
                matrix.dtype.itemsize,
                71 / (1 << 30),  # 71 byte in GB. Slightly less than two rows.
            ),
            2,
        )
        self.assertEqual(
            calc_chunk_size_from_memory_requirement(
                matrix.shape[0],
                matrix.shape[1],
                matrix.dtype.itemsize,
                72 / (1 << 30),  # 72 byte in GB. Exact size of matrix.
            ),
            3,
        )
        self.assertEqual(
            calc_chunk_size_from_memory_requirement(
                matrix.shape[0],
                matrix.shape[1],
                matrix.dtype.itemsize,
                73 / (1 << 30),  # 73 byte in GB. Slightly more than full matrix.
            ),
            3,
        )

        # test memory cutoff larger than matrix: chunk size contains all rows
        self.assertEqual(
            calc_chunk_size_from_memory_requirement(
                matrix.shape[0], matrix.shape[1], matrix.dtype.itemsize, 0.5
            ),  # 500mb
            3,
        )



"""Utility functions for visualization of molecules and their explanations."""

import io
from typing import Sequence

import numpy as np
import numpy.typing as npt
from matplotlib import pyplot as plt
from matplotlib.colors import Colormap, ListedColormap
from matplotlib.pyplot import get_cmap
from PIL import Image
from rdkit import Chem

# red green blue alpha tuple
RGBAtuple = tuple[float, float, float, float]


def get_mol_lims(mol: Chem.Mol) -> tuple[tuple[float, float], tuple[float, float]]:
    """Return the extent of the molecule.

    x- and y-coordinates of all atoms in the molecule are accessed, returning min- and max-values for both axes.

    Parameters
    ----------
    mol: Chem.Mol
        RDKit Molecule object of which the limits are determined.

    Returns
    -------
    tuple[tuple[float, float], tuple[float, float]]
        Limits of the molecule.
    """
    coords_list = []
    conf = mol.GetConformer(0)
    for i, _ in enumerate(mol.GetAtoms()):
        pos = conf.GetAtomPosition(i)
        coords_list.append((pos.x, pos.y))
    coords: npt.NDArray[np.float64] = np.array(coords_list)
    min_p = np.min(coords, axis=0)
    max_p = np.max(coords, axis=0)
    x_lim = min_p[0], max_p[0]
    y_lim = min_p[1], max_p[1]
    return x_lim, y_lim


def pad(
    lim: Sequence[float] | npt.NDArray[np.float64], ratio: float
) -> tuple[float, float]:
    """Take a 2-dimensional vector and adds len(vector) * ratio / 2 to each side and returns obtained vector.

    Parameters
    ----------
    lim: Sequence[float] | npt.NDArray[np.float64]
        Limits which are extended.
    ratio: float
        factor by which the limits are extended.

    Returns
    -------
    List[float, float]
        Extended limits
    """
    diff = max(lim) - min(lim)
    diff *= ratio / 2
    return lim[0] - diff, lim[1] + diff


def get_color_map_from_input(
    color: str | Colormap | tuple[RGBAtuple, RGBAtuple, RGBAtuple] | None,
) -> Colormap:
    """Get a colormap from a user defined color scheme.

    Parameters
    ----------
    color: str | Colormap | tuple[RGBAtuple, RGBAtuple, RGBAtuple] | None
        The color scheme.

    Returns
    -------
    Colormap
        The colormap.
    """
    # read user definer color scheme as ColorMap
    if color is None:
        coolwarm = (
            (1.0, 0.25, 0.0, 0.5),
            (1.0, 1.0, 1.0, 0.5),
            (0.017, 0.50, 0.850, 0.5),
        )
        color = coolwarm
    if isinstance(color, Colormap):
        color_map = color
    elif isinstance(color, tuple):
        color_map = color_tuple_to_colormap(color)  # type: ignore
    elif isinstance(color, str):
        color_map = get_cmap(color)
    else:
        raise ValueError("Color must be a tuple, string or ColorMap.")
    return color_map


def color_tuple_to_colormap(
    color_tuple: tuple[RGBAtuple, RGBAtuple, RGBAtuple],
) -> Colormap:
    """Convert a color tuple to a colormap.

    Parameters
    ----------
    color_tuple: tuple[RGBAtuple, RGBAtuple, RGBAtuple]
        The color tuple.

    Returns
    -------
    Colormap
        The colormap (a matplotlib data structure).
    """
    if len(color_tuple) != 3:
        raise ValueError("Color tuple must have 3 elements")

    # Definition of color
    col1, col2, col3 = map(np.array, color_tuple)

    # Creating linear gradient for color mixing
    linspace = np.linspace(0, 1, int(128))
    linspace4d = np.vstack([linspace] * 4).T

    # interpolating values for 0 to 0.5 by mixing purple and white
    zero_to_half = linspace4d * col2 + (1 - linspace4d) * col3
    # interpolating values for 0.5 to 1 by mixing white and yellow
    half_to_one = col1 * linspace4d + col2 * (1 - linspace4d)

    # Creating new colormap from
    color_map = ListedColormap(np.vstack([zero_to_half, half_to_one]))
    return color_map


def to_png(data: bytes) -> Image.Image:
    """Show a PNG image from a byte stream.

    Parameters
    ----------
    data: bytes
        The image data.

    Returns
    -------
    Image
        The image.
    """
    bio = io.BytesIO(data)
    img = Image.open(bio)
    return img


def plt_to_pil(figure: plt.Figure) -> Image.Image:
    """Convert a matplotlib figure to a PIL image.

    Parameters
    ----------
    figure: plt.Figure
        The figure.

    Returns
    -------
    Image
        The image.
    """
    bio = io.BytesIO()
    figure.savefig(bio, format="png")
    bio.seek(0)
    img = Image.open(bio)
    return img


def get_atom_coords_of_bond(
    bond: Chem.Bond, conf: Chem.Conformer
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """Get the two atom coordinates of a bond in the conformation.

    Parameters
    ----------
    bond: Chem.Bond
        The bond.
    conf: Chem.Conformer
        The conformation.

    Returns
    -------
    tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]
        The atom coordinates.
    """
    a1 = bond.GetBeginAtom().GetIdx()
    a1_pos = conf.GetAtomPosition(a1)
    a1_coords = np.array([a1_pos.x, a1_pos.y])

    a2 = bond.GetEndAtom().GetIdx()
    a2_pos = conf.GetAtomPosition(a2)
    a2_coords = np.array([a2_pos.x, a2_pos.y])

    return a1_coords, a2_coords


def calc_present_and_absent_shap_contributions(
    feature_vector: npt.NDArray[np.float64], feature_weights: npt.NDArray[np.float64]
) -> tuple[float, float]:
    """Get the sum of present and absent SHAP values.

    Parameters
    ----------
    feature_vector: npt.NDArray[np.float64]
        The feature vector.
    feature_weights: npt.NDArray[np.float64]
        The feature weights.

    Raises
    ------
    ValueError
        If the feature vector is not binary.

    Returns
    -------
    tuple[float, float]
        The sum of present and absent SHAP values.
    """
    if feature_vector.max() > 1 or feature_vector.min() < 0:
        raise ValueError(
            "Feature vector must be binary. Alternatively, use the structure_heatmap function instead."
        )

    # determine present/absent features using the binary feature vector
    present_shap = feature_weights * feature_vector
    absent_shap = feature_weights * (1 - feature_vector)
    sum_present_shap = sum(present_shap)
    sum_absent_shap = sum(absent_shap)

    return sum_present_shap, sum_absent_shap



"""Init file for estimators."""

from molpipeline.estimators.connected_component_clustering import (
    ConnectedComponentClustering,
)
from molpipeline.estimators.leader_picker_clustering import LeaderPickerClustering
from molpipeline.estimators.murcko_scaffold_clustering import MurckoScaffoldClustering
from molpipeline.estimators.nearest_neighbor import NamedNearestNeighbors
from molpipeline.estimators.similarity_transformation import TanimotoToTraining

__all__ = [
    "ConnectedComponentClustering",
    "LeaderPickerClustering",
    "MurckoScaffoldClustering",
    "NamedNearestNeighbors",
    "TanimotoToTraining",
]



"""Test the mol2scaffold module."""

from typing import Any
from unittest import TestCase

from molpipeline import Pipeline
from molpipeline.any2mol import AutoToMol
from molpipeline.mol2any import MolToSmiles
from molpipeline.mol2mol.scaffolds import MakeScaffoldGeneric, MurckoScaffold


class TestMurckoScaffold(TestCase):
    """Test the MurckoScaffold class."""

    def test_murcko_scaffold_generation_pipeline(self) -> None:
        """Test the scaffold generation."""
        scaffold_pipeline = Pipeline(
            steps=[
                ("smiles_to_mol", AutoToMol()),
                ("murcko_scaffold", MurckoScaffold()),
                ("scaffold_to_smiles", MolToSmiles()),
            ]
        )
        smiles_list = ["Cc1ccc(=O)[nH]c1", "O=CC1CCC(c2ccccc2)CC1", "CCC"]
        expected_scaffold_list = ["O=c1cccc[nH]1", "c1ccc(C2CCCCC2)cc1", ""]

        scaffold_list = scaffold_pipeline.transform(smiles_list)
        self.assertListEqual(expected_scaffold_list, scaffold_list)


class TestMakeScaffoldGeneric(TestCase):
    """Test the MakeScaffoldGeneric class."""

    def setUp(self) -> None:
        """Set up the pipeline and common variables."""
        self.generic_scaffold_pipeline = Pipeline(
            steps=[
                ("smiles_to_mol", AutoToMol()),
                ("murcko_scaffold", MurckoScaffold()),
                ("make_scaffold_generic", MakeScaffoldGeneric()),
                ("scaffold_to_smiles", MolToSmiles()),
            ]
        )
        self.smiles_list = ["Cc1ccc(=O)[nH]c1", "O=CC1CCC(c2ccccc2)CC1", "CCC"]

    def check_generic_scaffold(
        self, params: dict[str, Any], expected_scaffold_list: list[str]
    ) -> None:
        """Helper function to set parameters and check the results.

        Parameters
        ----------
        params: dict[str, Any]
            Parameters to set for the pipeline.
        expected_scaffold_list: list[str]
            Expected output of the pipeline.
        """
        self.generic_scaffold_pipeline.set_params(**params)
        generic_scaffold_list = self.generic_scaffold_pipeline.transform(
            self.smiles_list
        )
        self.assertListEqual(expected_scaffold_list, generic_scaffold_list)

    def test_generic_scaffold_generation_pipeline(self) -> None:
        """Test the generic scaffold generation."""
        self.check_generic_scaffold(
            params={}, expected_scaffold_list=["CC1CCCCC1", "C1CCC(C2CCCCC2)CC1", ""]
        )

        # Test the generic scaffold generation with generic atoms
        self.check_generic_scaffold(
            params={"make_scaffold_generic__generic_atoms": True},
            expected_scaffold_list=["**1*****1", "*1***(*2*****2)**1", ""],
        )

        # Test the generic scaffold generation with generic bonds
        self.check_generic_scaffold(
            params={
                "make_scaffold_generic__generic_atoms": False,
                "make_scaffold_generic__generic_bonds": True,
            },
            expected_scaffold_list=[
                "C~C1~C~C~C~C~C~1",
                "C1~C~C~C(~C2~C~C~C~C~C~2)~C~C~1",
                "",
            ],
        )

        # Test the generic scaffold generation with generic atoms and bonds
        self.check_generic_scaffold(
            params={
                "make_scaffold_generic__generic_atoms": True,
                "make_scaffold_generic__generic_bonds": True,
            },
            expected_scaffold_list=[
                "*~*1~*~*~*~*~*~1",
                "*1~*~*~*(~*2~*~*~*~*~*~2)~*~*~1",
                "",
            ],
        )



"""Wrapper for Chemprop to make it compatible with scikit-learn."""

from typing import Any, Sequence

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

import numpy as np
import numpy.typing as npt
from loguru import logger
from sklearn.base import clone
from sklearn.utils._tags import (
    ClassifierTags,
    RegressorTags,
    Tags,
)
from sklearn.utils.metaestimators import available_if

try:
    from chemprop.data import MoleculeDataset, build_dataloader
    from chemprop.nn.predictors import BinaryClassificationFFNBase
    from lightning import pytorch as pl
except ImportError as error:
    logger.error(
        "Chemprop is not installed. Please install it using `pip install chemprop`."
    )
    logger.info(error)


from molpipeline.estimators.chemprop.abstract import ABCChemprop
from molpipeline.estimators.chemprop.component_wrapper import (
    MPNN,
    BinaryClassificationFFN,
    BondMessagePassing,
    MulticlassClassificationFFN,
    RegressionFFN,
    SumAggregation,
)
from molpipeline.estimators.chemprop.neural_fingerprint import ChempropNeuralFP


class ChempropModel(ABCChemprop):
    """Wrap Chemprop in a sklearn like Estimator."""

    _classes_: npt.NDArray[np.int_] | None

    def __init__(
        self,
        model: MPNN,
        lightning_trainer: pl.Trainer | None = None,
        batch_size: int = 64,
        n_jobs: int = 1,
        **kwargs: Any,
    ) -> None:
        """Initialize the chemprop abstract model.

        Parameters
        ----------
        model : MPNN
            The chemprop model to wrap.
        lightning_trainer : pl.Trainer, optional
            The lightning trainer to use, by default None
        batch_size : int, optional (default=64)
            The batch size to use.
        n_jobs : int, optional (default=1)
            The number of jobs to use.
        kwargs : Any
            Parameters set using `set_params`.
            Can be used to modify components of the model.
        """
        super().__init__(
            model=model,
            lightning_trainer=lightning_trainer,
            batch_size=batch_size,
            n_jobs=n_jobs,
            **kwargs,
        )
        self._classes_ = None

    @property
    def classes_(self) -> npt.NDArray[np.int_]:
        """Return the classes."""
        if not self._is_classifier():
            raise ValueError("Model is not a classifier.")
        if self._classes_ is None:
            raise ValueError("Classes are not set.")
        return self._classes_

    @property
    def _estimator_type(self) -> str:
        """Return the estimator type."""
        if self._is_classifier():
            return "classifier"
        return "regressor"

    def _is_binary_classifier(self) -> bool:
        """Check if the model is a binary classifier.

        Returns
        -------
        bool
            True if the model is a binary classifier, False otherwise.
        """
        if isinstance(self.model.predictor, BinaryClassificationFFNBase):
            return True
        return False

    def _is_multiclass_classifier(self) -> bool:
        """Check if the model is a multiclass classifier.

        Returns
        -------
        bool
            True if the model is a multiclass classifier, False otherwise.
        """
        if isinstance(self.model.predictor, MulticlassClassificationFFN):
            return True
        return False

    def _is_classifier(self) -> bool:
        """Check if the model is a classifier.

        Returns
        -------
        bool
            True if the model is a classifier, False otherwise.
        """
        return self._is_binary_classifier() or self._is_multiclass_classifier()

    def __sklearn_tags__(self) -> Tags:
        """Return the sklearn tags.

        Returns
        -------
        Tags
            The sklearn tags for the model.
        """
        tags = super().__sklearn_tags__()
        if self._is_classifier():
            tags.estimator_type = "classifier"
            tags.classifier_tags = ClassifierTags()
        else:
            tags.estimator_type = "regressor"
            tags.regressor_tags = RegressorTags()
        tags.target_tags.required = True
        return tags

    def _predict(
        self, X: MoleculeDataset  # pylint: disable=invalid-name
    ) -> npt.NDArray[np.float64]:
        """Predict the labels.

        Parameters
        ----------
        X : MoleculeDataset
            The input data.

        Returns
        -------
        npt.NDArray[np.float64]
            The predictions for the input data.
        """
        self.model.eval()
        test_data = build_dataloader(X, num_workers=self.n_jobs, shuffle=False)
        predictions = self.lightning_trainer.predict(self.model, test_data)
        prediction_array = np.vstack(predictions)  # type: ignore
        prediction_array = prediction_array.squeeze(axis=1)
        # Check if the predictions have the same length as the input dataset
        if prediction_array.shape[0] != len(X):
            raise AssertionError(
                "Predictions should have the same length as the input dataset."
            )

        # If the model is a binary classifier, return the probability of the positive class
        if self._is_binary_classifier():
            if prediction_array.ndim != 1:
                raise ValueError(
                    "Binary classification model should output a single probability."
                )
        return prediction_array

    def fit(
        self,
        X: MoleculeDataset,
        y: Sequence[int | float] | npt.NDArray[np.int_ | np.float64],
    ) -> Self:
        """Fit the model to the data.

        Parameters
        ----------
        X : MoleculeDataset
            The input data.
        y : Sequence[int | float] | npt.NDArray[np.int_ | np.float64]
            The target data.

        Returns
        -------
        Self
            The fitted model.
        """
        if self._is_classifier():
            self._classes_ = np.unique(y)
        return super().fit(X, y)

    def predict(
        self, X: MoleculeDataset  # pylint: disable=invalid-name
    ) -> npt.NDArray[np.float64]:
        """Predict the output.

        Parameters
        ----------
        X : MoleculeDataset
            The input data.

        Returns
        -------
        npt.NDArray[np.float64]
            The predictions for the input data.
        """
        predictions = self._predict(X)
        if self._is_binary_classifier():
            pred = np.zeros(len(predictions))
            pred[predictions > 0.5] = 1
            return pred

        if self._is_multiclass_classifier():
            return np.argmax(predictions, axis=1)

        return predictions

    @available_if(_is_classifier)
    def predict_proba(
        self, X: MoleculeDataset  # pylint: disable=invalid-name
    ) -> npt.NDArray[np.float64]:
        """Predict the probabilities.

        Parameters
        ----------
        X : MoleculeDataset
            The input data.

        Returns
        -------
        npt.NDArray[np.float64]
            The probabilities of the input data.
        """
        if self._is_binary_classifier():
            proba_class_1 = self._predict(X)
            return np.vstack([1 - proba_class_1, proba_class_1]).T
        return self._predict(X)

    def to_encoder(self) -> ChempropNeuralFP:
        """Return the encoder for the model.

        Returns
        -------
        ChempropNeuralFP
            The encoder for the model.
        """
        return ChempropNeuralFP(
            model=clone(self.model),  # type: ignore
            lightning_trainer=self.lightning_trainer,
            batch_size=self.batch_size,
            n_jobs=self.n_jobs,
            disable_fitting=True,
        )


class ChempropClassifier(ChempropModel):
    """Chemprop model with default parameters for binary classification tasks."""

    def __init__(
        self,
        model: MPNN | None = None,
        lightning_trainer: pl.Trainer | None = None,
        batch_size: int = 64,
        n_jobs: int = 1,
        **kwargs: Any,
    ) -> None:
        """Initialize the chemprop classifier model.

        Parameters
        ----------
        model : MPNN | None, optional
            The chemprop model to wrap. If None, a default model will be used.
        lightning_trainer : pl.Trainer, optional
            The lightning trainer to use, by default None
        batch_size : int, optional (default=64)
            The batch size to use.
        n_jobs : int, optional (default=1)
            The number of jobs to use.
        kwargs : Any
            Parameters set using `set_params`.
            Can be used to modify components of the model.
        """
        if model is None:
            bond_encoder = BondMessagePassing()
            agg = SumAggregation()
            predictor = BinaryClassificationFFN()
            model = MPNN(message_passing=bond_encoder, agg=agg, predictor=predictor)
        super().__init__(
            model=model,
            lightning_trainer=lightning_trainer,
            batch_size=batch_size,
            n_jobs=n_jobs,
            **kwargs,
        )
        if not self._is_binary_classifier():
            raise ValueError("ChempropClassifier should be a binary classifier.")

    def set_params(self, **params: Any) -> Self:
        """Set the parameters of the model and check if it is a binary classifier.

        Parameters
        ----------
        **params
            The parameters to set.

        Returns
        -------
        Self
            The model with the new parameters.
        """
        super().set_params(**params)
        if not self._is_binary_classifier():
            raise ValueError("ChempropClassifier should be a binary classifier.")
        return self


class ChempropRegressor(ChempropModel):
    """Chemprop model with default parameters for regression tasks."""

    def __init__(
        self,
        model: MPNN | None = None,
        lightning_trainer: pl.Trainer | None = None,
        batch_size: int = 64,
        n_jobs: int = 1,
        **kwargs: Any,
    ) -> None:
        """Initialize the chemprop regressor model.

        Parameters
        ----------
        model : MPNN | None, optional
            The chemprop model to wrap. If None, a default model will be used.
        lightning_trainer : pl.Trainer, optional
            The lightning trainer to use, by default None
        batch_size : int, optional (default=64)
            The batch size to use.
        n_jobs : int, optional (default=1)
            The number of jobs to use.
        kwargs : Any
            Parameters set using `set_params`.
            Can be used to modify components of the model.
        """
        if model is None:
            bond_encoder = BondMessagePassing()
            agg = SumAggregation()
            predictor = RegressionFFN()
            model = MPNN(message_passing=bond_encoder, agg=agg, predictor=predictor)
        super().__init__(
            model=model,
            lightning_trainer=lightning_trainer,
            batch_size=batch_size,
            n_jobs=n_jobs,
            **kwargs,
        )


class ChempropMulticlassClassifier(ChempropModel):
    """Chemprop model with default parameters for multiclass classification tasks."""

    def __init__(
        self,
        n_classes: int,
        model: MPNN | None = None,
        lightning_trainer: pl.Trainer | None = None,
        batch_size: int = 64,
        n_jobs: int = 1,
        **kwargs: Any,
    ) -> None:
        """Initialize the chemprop multiclass model.

        Parameters
        ----------
        n_classes : int
            The number of classes for the classifier.
        model : MPNN | None, optional
            The chemprop model to wrap. If None, a default model will be used.
        lightning_trainer : pl.Trainer, optional
            The lightning trainer to use, by default None
        batch_size : int, optional (default=64)
            The batch size to use.
        n_jobs : int, optional (default=1)
            The number of jobs to use.
        kwargs : Any
            Parameters set using `set_params`.
            Can be used to modify components of the model.

        Raises
        ------
        AttributeError
            If the passed model.predictor does not have an attribute n_classes.
        ValueError
            If the number of classes in the predictor does not match the number of classes given as attribute.
        """
        if model is None:
            bond_encoder = BondMessagePassing()
            agg = SumAggregation()
            predictor = MulticlassClassificationFFN(n_classes=n_classes)
            model = MPNN(message_passing=bond_encoder, agg=agg, predictor=predictor)
        if not hasattr(model.predictor, "n_classes"):
            raise AttributeError(
                "The predictor does not have an attribute n_classes. Please use a MulticlassClassificationFFN predictor or define n_classes."
            )
        if n_classes != model.predictor.n_classes:
            raise ValueError(
                "The number of classes in the predictor does not match the number of classes."
            )
        super().__init__(
            model=model,
            lightning_trainer=lightning_trainer,
            batch_size=batch_size,
            n_jobs=n_jobs,
            **kwargs,
        )
        self._is_valid_multiclass_classifier()

    @property
    def n_classes(self) -> int:
        """Return the number of classes."""
        return self.model.predictor.n_classes

    @n_classes.setter
    def n_classes(self, n_classes: int) -> None:
        """Set the number of classes.

        Parameters
        ----------
        n_classes : int
            number of classes
        """
        self.model.predictor.n_classes = n_classes
        self.model.reinitialize_network()

    def set_params(self, **params: Any) -> Self:
        """Set the parameters of the model and check if it is a multiclass classifier.

        Parameters
        ----------
        **params
            The parameters to set.

        Returns
        -------
        Self
            The model with the new parameters.
        """
        super().set_params(**params)
        if not self._is_valid_multiclass_classifier():
            raise ValueError(
                "The model's predictor or the number of classes are invalid. Use a multiclass predictor and more than 2 classes."
            )
        return self

    def fit(
        self,
        X: MoleculeDataset,
        y: Sequence[int | float] | npt.NDArray[np.int_ | np.float64],
    ) -> Self:
        """Fit the model to the data.

        Parameters
        ----------
        X : MoleculeDataset
            The input data.
        y : Sequence[int | float] | npt.NDArray[np.int_ | np.float64]
            The target data.

        Returns
        -------
        Self
            The fitted model.
        """
        self._check_correct_input(y)
        return super().fit(X, y)

    def _check_correct_input(
        self, y: Sequence[int | float] | npt.NDArray[np.int_ | np.float64]
    ) -> None:
        """Check if the input for the multi-class classifier is correct.

        Parameters
        ----------
        y : Sequence[int | float] | npt.NDArray[np.int_ | np.float64]
            Indended classes for the dataset

        Raises
        ------
        ValueError
            If the classes found in y are not matching n_classes or if the class labels do not start from 0 to n_classes-1.
        """
        unique_y = np.unique(y)
        log = []
        if self.n_classes != len(unique_y):
            log.append(
                f"Given number of classes in init (n_classes) does not match the number of unique classes (found {unique_y}) in the target data."
            )
        if sorted(unique_y) != list(range(self.n_classes)):
            err = f"Classes need to be in the range from 0 to {self.n_classes-1}. Found {unique_y}. Please correct the input data accordingly."
            log.append(err)
        if log:
            raise ValueError("\n".join(log))

    def _is_valid_multiclass_classifier(self) -> bool:
        """Check if a multiclass classifier is valid. Model FFN needs to be of the correct class and model needs to have more than 2 classes.

        Returns
        -------
        bool
            True if is a valid multiclass classifier, False otherwise.
        """
        has_correct_model = isinstance(
            self.model.predictor, MulticlassClassificationFFN
        )
        has_classes = self.n_classes > 2
        return has_correct_model and has_classes



"""Wrapper for Chemprop GraphFeaturizer."""

from dataclasses import InitVar
from typing import Any

try:
    from typing import Self  # type: ignore[attr-defined]
except ImportError:
    from typing_extensions import Self

from chemprop.featurizers.molgraph import (
    SimpleMoleculeMolGraphFeaturizer as _SimpleMoleculeMolGraphFeaturizer,
)


class SimpleMoleculeMolGraphFeaturizer(_SimpleMoleculeMolGraphFeaturizer):
    """Wrapper for Chemprop SimpleMoleculeMolGraphFeaturizer."""

    extra_atom_fdim: InitVar[int]
    extra_bond_fdim: InitVar[int]

    def get_params(
        self, deep: bool = True  # pylint: disable=unused-argument
    ) -> dict[str, InitVar[int]]:
        """Get parameters for the featurizer.

        Parameters
        ----------
        deep: bool, optional (default=True)
            Used for compatibility with scikit-learn.

        Returns
        -------
        dict[str, int]
            Parameters of the featurizer.
        """
        return {}

    def set_params(self, **parameters: Any) -> Self:  # pylint: disable=unused-argument
        """Set the parameters of the featurizer.

        Parameters
        ----------
        parameters: Any
            Parameters to set. Only used for compatibility with scikit-learn.

        Returns
        -------
        Self
            This featurizer with the parameters set.
        """
        return self



"""Module for accessing the parameters of a lightning trainer."""

from pathlib import Path
from typing import Any

try:
    import lightning as pl
    import torch
    from lightning.pytorch.accelerators import CPUAccelerator, CUDAAccelerator
    from lightning.pytorch.callbacks import ModelSummary, ProgressBar, Timer
    from lightning.pytorch.trainer.connectors.accelerator_connector import Accelerator
except ImportError:
    pass


def get_enable_progress_bar(trainer: pl.Trainer) -> bool:
    """Get if the progress bar is enabled in the lightning trainer.

    Parameters
    ----------
    trainer : pl.Trainer
        The lightning trainer.

    Returns
    -------
    bool
        If the progress bar is enabled in the lightning trainer.
    """
    for callback in trainer.callbacks:  # type: ignore[attr-defined]
        if isinstance(callback, ProgressBar):
            return True
    return False


def get_device(trainer: pl.Trainer) -> str | Accelerator:
    """Get the device used by the lightning trainer.

    Parameters
    ----------
    trainer : pl.Trainer
        The lightning trainer.

    Returns
    -------
    str
        The device used by the lightning trainer.
    """
    devices: str | Accelerator
    if isinstance(trainer.accelerator, CPUAccelerator):
        devices = "cpu"
    elif isinstance(trainer.accelerator, CUDAAccelerator):
        devices = "gpu"
    else:
        devices = trainer.accelerator
    return devices


TRAINER_DEFAULT_PARAMS = {
    "strategy": "auto",
    "limit_predict_batches": 1.0,
    "fast_dev_run": False,
    "min_steps": None,
    "accumulate_grad_batches": 1,
    "use_distributed_sampler": True,
    "devices": "auto",
    "check_val_every_n_epoch": 1,
    "enable_progress_bar": True,
    "max_epochs": None,
    "max_time": None,
    "val_check_interval": 1.0,
    "log_every_n_steps": 50,
    "min_epochs": None,
    "gradient_clip_algorithm": None,
    "max_steps": -1,
    "limit_val_batches": 1.0,
    "gradient_clip_val": None,
    "inference_mode": True,
    "enable_model_summary": None,
    "limit_test_batches": 1.0,
    "reload_dataloaders_every_n_epochs": 0,
    "deterministic": False,
    "logger": None,
    "overfit_batches": 0.0,
    "precision": "32-true",
    "benchmark": False,
    "num_sanity_val_steps": 2,
    "enable_checkpointing": None,
    "limit_train_batches": 1.0,
    "barebones": False,
    "detect_anomaly": False,
    "num_nodes": 1,
}


def get_trainer_path(trainer: pl.Trainer) -> str | None:
    """Get the path of the lightning trainer.

    Parameters
    ----------
    trainer : pl.Trainer
        The lightning trainer.

    Returns
    -------
    str | None
        The path of the lightning trainer.
        None if the path is the current path.
    """
    curr_path = str(Path(".").resolve())
    trainer_path: str | None = trainer.default_root_dir
    if trainer_path == curr_path:
        trainer_path = None
    return trainer_path


def get_params_trainer(trainer: pl.Trainer) -> dict[str, Any]:
    """Get the parameters of the lightning trainer.

    Parameters
    ----------
    trainer : pl.Trainer
        The lightning trainer.

    Returns
    -------
    dict[str, Any]
        The parameters of the lightning trainer.
    """
    if trainer.callbacks and isinstance(trainer.callbacks[-1], Timer):  # type: ignore[attr-defined]
        max_time = trainer.callbacks[-1].duration  # type: ignore[attr-defined]
    else:
        max_time = None

    for callback in trainer.callbacks:  # type: ignore[attr-defined]
        if isinstance(callback, ModelSummary):
            enable_progress_model_summary = True
            break
    else:
        enable_progress_model_summary = False

    trainer_dict = {
        "accelerator": get_device(trainer),
        "strategy": "auto",  # trainer.strategy, # collides with accelerator
        "devices": "auto",  # trainer._accelerator_connector._devices_flag does not really work
        "num_nodes": trainer.num_nodes,
        "precision": trainer.precision,
        "logger": trainer.logger,
        # "callbacks": trainer.callbacks,  # type: ignore[attr-defined]
        "fast_dev_run": trainer.fast_dev_run,  # type: ignore[attr-defined]
        "max_epochs": trainer.max_epochs,
        "min_epochs": trainer.min_epochs,
        "max_steps": trainer.max_steps,
        "min_steps": trainer.min_steps,
        "max_time": max_time,
        "limit_train_batches": trainer.limit_train_batches,
        "limit_val_batches": trainer.limit_val_batches,
        "limit_test_batches": trainer.limit_test_batches,
        "limit_predict_batches": trainer.limit_predict_batches,
        "overfit_batches": trainer.overfit_batches,  # type: ignore[attr-defined]
        "val_check_interval": trainer.val_check_interval,
        "check_val_every_n_epoch": trainer.check_val_every_n_epoch,
        "num_sanity_val_steps": trainer.num_sanity_val_steps,
        "log_every_n_steps": trainer.log_every_n_steps,  # type: ignore[attr-defined]
        "enable_checkpointing": bool(trainer.checkpoint_callbacks),
        "enable_progress_bar": get_enable_progress_bar(trainer),
        "enable_model_summary": enable_progress_model_summary,
        "accumulate_grad_batches": trainer.accumulate_grad_batches,
        "gradient_clip_val": trainer.gradient_clip_val,
        "gradient_clip_algorithm": trainer.gradient_clip_algorithm,
        "deterministic": torch.are_deterministic_algorithms_enabled(),
        "benchmark": torch.backends.cudnn.benchmark,
        "inference_mode": trainer.predict_loop.inference_mode,
        "use_distributed_sampler": trainer._accelerator_connector.use_distributed_sampler,  # pylint: disable=protected-access
        # "profiler": trainer.profiler,  # type: ignore[attr-defined]
        "detect_anomaly": trainer._detect_anomaly,  # pylint: disable=protected-access
        "barebones": trainer.barebones,
        # "plugins": trainer.plugins,  # can not be exctracted
        # "sync_batchnorm": trainer._accelerator_connector.sync_batchnorm,  # plugin related
        "reload_dataloaders_every_n_epochs": trainer.reload_dataloaders_every_n_epochs,  # type: ignore[attr-defined]
        "default_root_dir": get_trainer_path(trainer),
    }
    return trainer_dict


def get_non_default_params_trainer(trainer: pl.Trainer) -> dict[str, Any]:
    """Get the parameters of the lightning trainer which are not default.

    Parameters
    ----------
    trainer : pl.Trainer
        The lightning trainer.

    Returns
    -------
    dict[str, Any]
        The parameters of the lightning trainer.
    """
    trainer_dict = get_params_trainer(trainer)
    non_default_values = {}
    for key, value in trainer_dict.items():
        if key not in TRAINER_DEFAULT_PARAMS:
            non_default_values[key] = value
        elif value != TRAINER_DEFAULT_PARAMS[key]:
            non_default_values[key] = value
    return non_default_values



"""Explainability module for the molpipeline package."""

from molpipeline.experimental.explainability.explainer import (
    SHAPKernelExplainer,
    SHAPTreeExplainer,
)
from molpipeline.experimental.explainability.explanation import (
    SHAPFeatureAndAtomExplanation,
    SHAPFeatureExplanation,
)
from molpipeline.experimental.explainability.visualization.visualization import (
    structure_heatmap,
    structure_heatmap_shap,
)

__all__ = [
    "SHAPTreeExplainer",
    "SHAPKernelExplainer",
    "SHAPFeatureExplanation",
    "SHAPFeatureAndAtomExplanation",
    "structure_heatmap",
    "structure_heatmap_shap",
]
